/*\
|*| ---------------------------------------------------------------------------
|*|	replacement for linux/drivers/char/console.c
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright:
|*|	  (C) 1991, 1992	Linus Torvalds
|*|	Partial Copyright:
|*|	  (C) 1995, 1996	The SCRDRV project and
|*|	  (C) 1995, 1996	Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*|	This code was derived from Linux v1.2.6 code, especially 
|*|		linux/drivers/char/console.c
|*|		linux/drivers/char/vt.c
|*|
|*|	Therefore additional copyrights and credits apply. See the original
|*|	files for more detail. Hope this is a complete VT102 implementation.
|*| ---------------------------------------------------------------------------
|*|	This module exports the console io functions:
|*|
|*|	long con_init(long)			- initialize console module
|*|	void console_print(const char *s)	- print string on fg_console
|*|	void scrollback(int lines)		- hardware scroll support
|*|	void scrollfront(int lines)		- hardware scroll support
|*|	int  change_console_tty_ops(struct tty_driver *new)
|*|						- exchange console driver fn's
|*|	void reset_console_tty_ops(void)	- undo driver fn's override
|*| ---------------------------------------------------------------------------
|*|   NOTE:	We sometimes disable interrupts while processing a request
|*|		(to put a word in video IO), but this will work even for 
|*|		keyboard interrupts. We know interrupts aren't enabled when
|*|		getting a keyboard interrupt, as we use trap-gates.
|*|		Hopefully all is well.
\*/

#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <linux/malloc.h>
#include <linux/major.h>

#include <asm/io.h>

#include <linux/kbd.h>
#include "consolemap.h"

#include <linux/console.h>
#include <linux/ascii.h>

static int console_refcount;
static struct tty_struct	*console_table[MAX_NR_CONSOLES];
static struct termios		*console_termios[MAX_NR_CONSOLES];
static struct termios		*console_termios_locked[MAX_NR_CONSOLES];
struct tty_driver		console_driver;

int fg_console;			/* current valid console */
int last_console;		/* last used console     */
struct kgi_vc *vcons[MAX_NR_CONSOLES];
struct kgi_vc *boot_console;

/*\
|*| ---------------------------------------------------------------------------
|*|	The terminal emulation code.
|*| ---------------------------------------------------------------------------
|*|	All these functions work on a data structure representing a virtual
|*|	console. However, it may actually be a 'real' console depending on fb
|*|	pointing either to video memory or just normal RAM.
|*|	So we have *really* independent virtual consoles and support for multi-
|*|	display machines. We don't care for VC switches here as this requires
|*|	knowledge of how to programm the graphics card.
|*|	Anyhow, as we have only one keyboard, only vcons[fg_console] will have
|*|	the keyboard and pointer focus.
|*| ---------------------------------------------------------------------------
\*/

#define set_kbd(X) set_vc_kbd_mode(vc_kbd+VC_NUM, X)
#define clr_kbd(X) clr_vc_kbd_mode(vc_kbd+VC_NUM, X)
#define	set_kbd_led(X)	set_vc_kbd_led(vc_kbd+VC_NUM, X)
#define	clr_kbd_led(X)	clr_vc_kbd_led(vc_kbd+VC_NUM, X)
#define is_kbd(X)  vc_kbd_mode(vc_kbd+VC_NUM, X)

/*\
|*|	Bitmap for codes <32. A bit of 1 indicates that the code corresponding
|*|	to that bit number invokes some special action (such as cursor movement)
|*|	and should not be displayed as a glyph unless the disp_ctrl mode is
|*|	explicitly enabled.
\*/
#define CTRL_ACTION 0xd00ff80

static u_char color_table[] =
{  0,  4,  2,  6,  1,  5,  3,  7,  8, 12, 10, 14,  9, 13, 11, 15 };

/*\
|*|	Generates sound of some count for some number of clock ticks
|*|	[count = 1193180 / frequency]
|*|
|*|	If count is 0, will turn off sound, else will turn it on for that time.
|*|	If ticks is 0, will return immediately, else will sleep for msec time,
|*|	then turn sound off.
|*|
|*|	We use the BEEP_TIMER vector since we're using the same method to
|*|	generate sound, and we'll overwrite any beep in progress. That may
|*|	be something to fix later, if we like.
|*|
|*|	We also return immediately, which is what was implied within the X
|*|	comments - KDMKTONE doesn't put the process to sleep.
\*/

static void
kd_nosound(unsigned long ignored)
{
  outb(inb_p(0x61)&0xFC, 0x61);		/* disable counter 2 */
  return;
}

static void
kd_mksound(unsigned int count, unsigned int ticks)
{
  static struct timer_list sound_timer = { NULL, NULL, 0, 0, kd_nosound };

  cli();
  del_timer(&sound_timer);

  if (count)
    {
      outb_p(inb_p(0x61)|3, 0x61);	/* enable counter 2 */

      outb_p(0xB6, 0x43);		/* cmd for counter 2, 2 byte write */

      outb_p(count & 0xff, 0x42);	/* select desired HZ */
      outb((count >> 8) & 0xff, 0x42);

      if (ticks)
	{
	  sound_timer.expires = ticks;
	  add_timer(&sound_timer);
	}
    }
   else
    kd_nosound(0);

  sti();
  return;
}

static void
gotoxy(struct kgi_vc *cons, int new_x, int new_y)
{
  /*\
  |*|	gotoxy() must verify all boundaries, because the arguments might also
  |*|	be negative. If the given position is out of bounds, the cursor is
  |*|	placed at the nearest margin.
  \*/

  int max_y;

  X = (new_x < 0) ? 0 : ((new_x < SIZE_X) ? new_x : SIZE_X - 1);

  if (ORIGIN_MODE) 
    {
      new_y += TOP;
      max_y = BOTTOM;
    }
   else
    max_y = SIZE_Y;

  Y = (new_y < 0) ? 0 : ((new_y < max_y) ? new_y : max_y - 1);

  POS = ORIGIN + Y*FB_SIZE_X + X;
  NEED_WRAP = 0;
}

static void
scrup(struct kgi_vc *cons, unsigned int t, unsigned int b)
{
  /*\
  |*|	scroll screen between t(op) and b(ottom) up one line. If the virtual
  |*|	vertical screen size does not equal the console vertical size 
  |*|	hardscrolling is used. As hardscrolling may depend on the graphics 
  |*|	hardware we call the VC-specific set_origin function (if there is one).
  \*/

  int hardscroll = 1;

  if ((b > SIZE_Y) || (t >= b))
    return;

  if ((FB_SIZE_Y == SIZE_Y) || !SET_ORIGIN)
    hardscroll = 0;
   else 
    if (t || (b != SIZE_Y))
      hardscroll = 0;

  if (hardscroll)
    {
       ORIGIN  += FB_SIZE_X;
       POS     += FB_SIZE_X;
       SCR_END += FB_SIZE_X;
       if (SCR_END > FB_END)
         {
            u_short *d = FB;
            u_short *s = FB + ORIGIN;
            unsigned int count = (SIZE_Y - 1)*FB_SIZE_X;

            while (count--)
              *(d++) = *(s++);

            count = FB_SIZE_X;
            while (count--)
              *(d++) = ERASE;
			
            SCR_END -= ORIGIN;
            POS     -= ORIGIN;
            ORIGIN   = 0;
            HAS_SCROLLED = 1;
         }
        else
         {
            u_short *d = FB + (SCR_END - FB_SIZE_X);
            unsigned int count = FB_SIZE_X;

            while (count--)
              *(d++) = ERASE;
         }
       __ORIGIN = ORIGIN;
       SET_ORIGIN(ORIGIN);
    }
   else
    {
       u_short *d = FB + (ORIGIN + FB_SIZE_X*t);
       u_short *s = FB + (ORIGIN + FB_SIZE_X*(t+1));

       unsigned int count = (b-t-1)*FB_SIZE_X;

       while (count--) 
         *(d++) = *(s++);

       count = FB_SIZE_X;
       while (count--)
         *(d++) = ERASE;
    }
}

static void
scrdown(struct kgi_vc *cons, unsigned int t, unsigned int b)
{
  /*\
  |*|	scroll screen between t(op) and b(ottom) down one line
  \*/

   u_short *d = FB + (ORIGIN + FB_SIZE_X*b);
   u_short *s = FB + (ORIGIN + FB_SIZE_X*(b-1));
   unsigned int count = (b-t-1)*FB_SIZE_X;

   if ((b > SIZE_Y) || (t >= b))
     return;

   while (count--)
     *(--d) = *(--s);

   count = FB_SIZE_X;
   while (count--)
     *(--d) = ERASE;
	
   HAS_SCROLLED = 1;
}

static void
lf(struct kgi_vc *cons)
{
  /*\ 
  |*|	don't scroll if above bottom or below scrolling region
  \*/

  if ((Y+1) == BOTTOM)
    scrup(cons, TOP, BOTTOM);
   else
    if (Y < (SIZE_Y-1))
      {
        Y++;
	POS += FB_SIZE_X;
      }
  NEED_WRAP = 0;
}

static void
ri(struct kgi_vc *cons)
{
  /*\
  |*|	don't scroll if below top or above scrolling region
  \*/

  if (Y == TOP)
    scrdown(cons, TOP, BOTTOM);
   else
    if (Y > 0)
      {
        Y--;
        POS -= FB_SIZE_Y;
      }

  NEED_WRAP = 0;
}

static inline void
cr(struct kgi_vc *cons)
{
   POS -= X;
   NEED_WRAP = X = 0;
}

static inline void
bs(struct kgi_vc *cons)
{
  if (X) 
    {
      POS--;
      X--;
      NEED_WRAP = 0;
    }
}

static inline void
del(struct kgi_vc *cons)
{
  /*\
  |*|	ignored, but can be implemented
  \*/
}

static void
csi_J(struct kgi_vc *cons, int vpar)
{
  unsigned int count;
  u_short *start;

  switch (vpar)				/* erase ... */
   {
     case 0:	count = SCR_END - POS;		/* from cursor to end of disp.*/
		start = FB + POS;
		break;

     case 1:	count = POS - ORIGIN + 1;	/* from start to cursor	      */
		start = FB + ORIGIN;
		break;

     case 2:	count = FB_SIZE_X*SIZE_Y;	/* whole display	      */
		start = FB + ORIGIN;
		break;

     default:	return;
   }

  while (count--)
   *(start++) = ERASE;

  NEED_WRAP = 0;
}

static void
csi_K(struct kgi_vc *cons, int vpar)
{
  unsigned int count;
  u_short *start;

  switch (vpar)
   {
     case 0:	count = FB_SIZE_X - X;	/* erase from cursor to end of line  */
		start = FB + POS;
		break;

     case 1:	start = FB + (POS - X);	/* erase from start of line to cursor */
		count = X + 1;
		break;

     case 2:	start = FB + (POS - X);	/* erase whole line */
		count = FB_SIZE_X;
		break;

     default:	return;
   }

  while (count--) 
   *(start++) = ERASE;
	
  NEED_WRAP = 0;
}

static void
csi_X(struct kgi_vc *cons, int vpar)
{
  unsigned int count;
  u_short *start;

  if (!vpar) vpar++;

  start = FB + POS;	/* erase the following vpar positions (not VT100 ???) */
  count = (vpar > SIZE_X - X) ? (FB_SIZE_X - X) : vpar;

  while (count--)
   *(start++) = ERASE;

  NEED_WRAP = 0;
}

static void
update_attr(struct kgi_vc *cons)
{
  ATTR = COLOR;

  if (CAN_DO_COLOR) 
    {
      if (UNDERLINE)
	ATTR = (ATTR & 0xf0) | ULCOLOR;
       else
        if (INTENSITY == FL_HALF)
	  ATTR = (ATTR & 0xf0) | HALFCOLOR;
    }
  if ((REVERSE || SCREEN_MODE) && !(REVERSE && SCREEN_MODE))
    ATTR = reverse_video_char(ATTR);
  if (BLINK)
    ATTR ^= 0x80;
  if (INTENSITY == FL_BOLD)
    ATTR ^= 0x08;
  if (!CAN_DO_COLOR)
    {
      if (UNDERLINE)
        ATTR = (ATTR & 0xf8) | 0x01;
       else
        if (INTENSITY == FL_HALF)
          ATTR = (ATTR & 0xf0) | 0x08;
    }
  ERASE = SCREEN_MODE 
               ? ((reverse_video_char(COLOR) << 8) | ' ')
               : ((COLOR << 8) | ' ');
}

static void
default_attr(struct kgi_vc *cons)
{
  ATTR_FLAGS = FL_NORMAL;
  COLOR = DEF_COLOR;
}

static void
csi_m(struct kgi_vc *cons)
{
  int i;

  for (i = 0; i <= NPAR; i++)
   switch (PAR[i])
    {
      case 0:	/* all attributes off */
		default_attr(cons);
		break;

      case 1:	ATTR_FLAGS &= ~FL_INTENSITY;
		ATTR_FLAGS |= FL_BOLD;
		break;

      case 2:	ATTR_FLAGS &= ~FL_INTENSITY;
		ATTR_FLAGS |= FL_HALF;
		break;

      case 4:	ATTR_FLAGS |= FL_UNDERLINE;
		break;

      case 5:	ATTR_FLAGS |= FL_BLINK;
		break;

      case 7:	ATTR_FLAGS |= FL_REVERSE;
		break;

      case 10:	/*\
		|*|	ANSI X3.64-1979 (SCO-ish?)
		|*|	Select primary font, don't display
		|*|	control chars if defined, don't set
		|*|	bit 8 on output.
		\*/

		TRANSLATE = set_translate(CHARSET == 0
						? G0_CHARSET
						: G1_CHARSET);
		MODE_FLAGS &= ~FL_DISP_CTRL;
		MODE_FLAGS &= ~FL_TOGGLE_META;
		break;

      case 11:	/*\
		|*|	ANSI X3.64-1979 (SCO-ish?)
		|*|	Select first alternate font, let's
		|*|	chars < 32 be displayed as ROM chars.
		\*/
		TRANSLATE = set_translate(NULL_MAP);
		MODE_FLAGS |= FL_DISP_CTRL;
		MODE_FLAGS &= ~FL_TOGGLE_META;
		break;

      case 12:	/*\
		|*|	ANSI X3.64-1979 (SCO-ish?)
		|*|	Select second alternate font, toggle
		|*|	high bit before displaying as ROM char.
		\*/
		TRANSLATE = set_translate(NULL_MAP);
		MODE_FLAGS |= FL_DISP_CTRL;
		MODE_FLAGS |= FL_TOGGLE_META;;
		break;

      case 21:
      case 22:	ATTR_FLAGS &= ~FL_INTENSITY;
		ATTR_FLAGS |= FL_NORMAL;
		break;

      case 24:	ATTR_FLAGS &= ~FL_UNDERLINE;
		break;

      case 25:	ATTR_FLAGS &= ~FL_BLINK;
		break;

      case 27:	ATTR_FLAGS &= ~FL_REVERSE;
		break;

      case 38:	/*\
		|*|	ANSI X3.64-1979 (SCO-ish?)
		|*|	Enables underscore, white foreground with white
		|*|	underscore (Linux - use default foreground).
		\*/

		COLOR = (DEF_COLOR & 0x0F) | BACKGROUND;
		ATTR_FLAGS |= FL_UNDERLINE;
		break;

      case 39:	/*\
		|*|	ANSI X3.64-1979 (SCO-ish?)
		|*|	Disable underline option. Reset color to default???
		|*|	It did this before:
		\*/

		COLOR = (DEF_COLOR & 0x0F) | BACKGROUND;
		ATTR_FLAGS &= ~FL_UNDERLINE;
		break;

      case 49:	COLOR = (DEF_COLOR & 0xF0) | FOREGROUND;
		break;

      default:	if ((PAR[i] >= 30) && (PAR[i] <= 37))
		  COLOR = color_table[PAR[i]-30] | BACKGROUND; 
		 else
		  if ((PAR[i] >= 40) && (PAR[i] <= 47))
		    COLOR = (color_table[PAR[i]-40]<<4) | FOREGROUND;
		break;
   }
  update_attr(cons);
}

static void
respond_string(const char *p, struct tty_struct *tty)
{
  while (*p) 
   {
     tty_insert_flip_char(tty, *p, 0);
     p++;
   }
  tty_schedule_flip(tty);
}

static void
cursor_report(struct kgi_vc *cons, struct tty_struct * tty)
{
  char buf[40];

  sprintf(buf, "\033[%hd;%hdR", Y + (ORIGIN_MODE ? TOP+1 : 1), X+1);
  respond_string(buf, tty);
}

static inline void
status_report(struct tty_struct * tty)
{
  respond_string("\033[0n", tty);	/* Terminal ok */
}

static inline void
respond_ID(struct tty_struct * tty)
{
  #define VT100ID	"\033[?1;2c"
  #define VT102ID	"\033[?6c"
  respond_string(VT102ID, tty);
}

#if 0
void mouse_report(struct tty_struct * tty, int butt, int mrx, int mry)
{
	char buf[8];

	sprintf(buf, "\033[M%c%c%c", (char)(' ' + butt), (char)('!' + mrx),
		(char)('!' + mry));
	respond_string(buf, tty);
}

/* invoked via ioctl(TIOCLINUX) */
int mouse_reporting(void)
{
	int currcons = fg_console;

	return report_mouse;
}


static inline unsigned short *screenpos(int currcons, int offset, int viewed)
{
	unsigned short *p = (unsigned short *)(origin + offset);
	if (viewed && currcons == fg_console)
		p -= (__real_origin - __origin);
	return p;
}

/* Note: inverting the screen twice should revert to the original state */
void invert_screen(int currcons, int offset, int count, int viewed)
{
	unsigned short *p;

	count /= 2;
	p = screenpos(currcons, offset, viewed);
	if (can_do_color)
		while (count--) {
			unsigned short old = scr_readw(p);
			scr_writew(reverse_video_short(old), p);
			p++;
		}
	else
		while (count--) {
			unsigned short old = scr_readw(p);
			scr_writew(old ^ (((old & 0x0700) == 0x0100)
					  ? 0x7000 : 0x7700), p);
			p++;
		}
}

/* used by selection: complement pointer position */
void complement_pos(int currcons, int offset)
{
	static unsigned short *p = NULL;
	static unsigned short old = 0;

	if (p)
		scr_writew(old, p);
	if (offset == -1)
		p = NULL;
	else {
		p = screenpos(currcons, offset, 1);
		old = scr_readw(p);
		scr_writew(old ^ 0x7700, p);
	}
}

/* used by selection */
unsigned short screen_word(int currcons, int offset, int viewed)
{
	return scr_readw(screenpos(currcons, offset, viewed));
}

/* used by vcs - note the word offset */
unsigned short *screen_pos(int currcons, int w_offset, int viewed)
{
	return screenpos(currcons, 2 * w_offset, viewed);
}

void getconsxy(int currcons, char *p)
{
	p[0] = x;
	p[1] = y;
}

void putconsxy(int currcons, char *p)
{
	gotoxy(currcons, p[0], p[1]);
	set_cursor(currcons);
}
#endif

static inline void
set_cursor(struct kgi_vc *cons)
{
  CURSOR.x = X;
  CURSOR.y = Y;
  if (CURSOR_MODE)
    {
      if (CURSOR.show)
        CURSOR.show(&CURSOR, POS);
    }
   else
    if (CURSOR.hide)
      CURSOR.hide(&CURSOR, POS);
}

static inline void
set_pointer(struct kgi_vc *cons)
{
  if (POINTER_MODE)
    {
      if (POINTER.show)
        POINTER.show(&POINTER, PPOS);
    }
   else
    if (POINTER.hide)
      POINTER.hide(&POINTER, PPOS);
}

static void set_mode(struct kgi_vc *cons, int on_off)
{
  int i;

  for (i = 0; i <= NPAR; i++)
  if (QUES)
    switch(PAR[i])
     {	/* DEC private modes set/reset */
	case 1:	if (on_off)	/* cursor keys send ^[Ox/^[[x */
		  set_kbd(VC_CKMODE);
		 else
		  clr_kbd(VC_CKMODE);
		break;

	case 3: /* change between 80/132 column mode. We just change the */
		/* logical size. No mode switch is implied. */
/* !!!		vc_resize(cons, SIZE_Y, on_off ? 132 : 80); */
		break;

	case 5:	if ( (!SCREEN_MODE) != (!on_off))	/* inverted screen */
		  {
		    MODE_FLAGS &= ~FL_SCREEN_MODE;
		    MODE_FLAGS |= on_off ? FL_SCREEN_MODE : 0;
/* !!!		    invert_screen(cons, 0, video_screen_size, 0); */
		    update_attr(cons);
		  }
		break;

	case 6: MODE_FLAGS &= ~FL_ORIGIN_MODE;	/* origin relative/absolute */
		MODE_FLAGS |= on_off ? FL_ORIGIN_MODE : 0;
		gotoxy(cons,0,0);
		break;

	case 7: MODE_FLAGS &= ~FL_AUTO_WRAP;	/* autowrap on/off */
		MODE_FLAGS |= on_off ? FL_AUTO_WRAP : 0;
		break;

	case 8:	if (on_off)			/* autorepeat on/off */
		  set_kbd(VC_REPEAT);
		 else
		  clr_kbd(VC_REPEAT);
		break;
#if 0
	case 9:
		report_mouse = on_off ? 1 : 0;
		break;
#endif
	case 25:
		MODE_FLAGS &= ~FL_CURSOR_MODE; /* Cursor on/off */
		MODE_FLAGS |= on_off ? FL_CURSOR_MODE : 0;
		set_cursor(cons);
		break;
#if 0
	case 1000:
		report_mouse = on_off ? 2 : 0;
		break;
#endif
     }
   else
    switch(PAR[i])
     {  /* ANSI modes set/reset */
	case 3:	MODE_FLAGS &= ~FL_DISP_CTRL;	/* monitor (display ctrls) */
		MODE_FLAGS |= on_off ? FL_DISP_CTRL : 0;
		break;

	case 4:	MODE_FLAGS &= ~FL_DISP_CTRL;	/* insert mode on/off */
		MODE_FLAGS |= on_off ? FL_INSERT_MODE : 0;
		break;

	case 20:
		if (on_off)			/* LF, ENTER = CRLF/LF */
		  set_kbd(VC_CRLF);
		 else
		  clr_kbd(VC_CRLF);
		break;
      }
}

static void
setterm_command(struct kgi_vc *cons)
{
  switch(PAR[0])
   {
     case 1:	if (CAN_DO_COLOR && (PAR[1] < 16))  /* set underline color */
		  {
		    ULCOLOR = color_table[PAR[1]];
		    if (UNDERLINE)
			update_attr(cons);
		  }
		break;

     case 2:	if (CAN_DO_COLOR && (PAR[1] < 16)) /* set half intensity color */
		  {
		    HALFCOLOR = color_table[PAR[1]];
		    if (INTENSITY == FL_HALF)
		      update_attr(cons);
		  }
		break;

     case 8:	DEF_COLOR = ATTR;		/* store colors as defaults */
		default_attr(cons);
		update_attr(cons);
		break;
#if 0
     case 9:	/* set blanking interval */
		blankinterval = ((par[1] < 60) ? par[1] : 60) * 60 * HZ;
		poke_blanked_console();
		break;
#endif
   }
}

static void
insert_char(struct kgi_vc *cons)
{
  unsigned int i = X;
  u_short tmp, old = ERASE;
  register u_short *p = FB + POS;

  while (i++ < SIZE_X)
   {
     tmp = *p;
     *(p++) = old;
     old = tmp;
   }
  NEED_WRAP = 0;
}

static void
insert_line(struct kgi_vc *cons)
{
  scrdown(cons, Y, BOTTOM);
  NEED_WRAP = 0;
}

static void
delete_char(struct kgi_vc *cons)
{
  unsigned int i = X;
  u_short *p = FB + POS;

  while (++i < SIZE_X)
   {
     *p = *(p+1);
     p++;
   }
  *p = ERASE;
  NEED_WRAP = 0;
}

static void
delete_line(struct kgi_vc *cons)
{
  scrup(cons,Y,BOTTOM);
  NEED_WRAP = 0;
}

static void
csi_at(struct kgi_vc *cons, unsigned int nr)
{
  nr = (SIZE_X < nr) ? FB_SIZE_X : (nr) ? nr : 1;

  while (nr--)
   insert_char(cons);
}

static void
csi_L(struct kgi_vc *cons, unsigned int nr)
{
  nr = (SIZE_Y < nr) ? SIZE_Y : (nr) ? nr : 1;

  while (nr--)
   insert_line(cons);
}

static void
csi_P(struct kgi_vc *cons, unsigned int nr)
{
  nr = (SIZE_X < nr) ? FB_SIZE_X : (nr) ? nr : 1;

  while (nr--)
   delete_char(cons);
}

static void
csi_M(struct kgi_vc *cons, unsigned int nr)
{
  nr = (nr > SIZE_Y) ? SIZE_Y : (nr) ? nr : 1; 

  while (nr--)
   delete_line(cons);
}

static void
save_cur(struct kgi_vc *cons)
{
	S_X		= X;
	S_Y		= Y;
	S_ATTR_FLAGS	= ATTR_FLAGS;
	S_CHARSET	= CHARSET;
	S_COLOR		= COLOR;
	S_G0_CHARSET	= G0_CHARSET;
	S_G1_CHARSET	= G1_CHARSET;
}

static void
restore_cur(struct kgi_vc *cons)
{
   gotoxy(cons, S_X, S_Y);
   ATTR_FLAGS	= S_ATTR_FLAGS;
   CHARSET	= S_CHARSET;
   COLOR	= S_COLOR;
   G0_CHARSET	= S_G0_CHARSET;
   G1_CHARSET	= S_G1_CHARSET;
   TRANSLATE	= set_translate(CHARSET ? G1_CHARSET : G0_CHARSET);
   update_attr(cons);
   NEED_WRAP	= 0;
}

enum { ESnormal, ESesc, ESsquare, ESgetpars, ESgotpars, ESfunckey, 
	EShash, ESsetG0, ESsetG1, ESpercent, ESignore };


void
reset_terminal(struct kgi_vc *cons, int do_clear)
{
  TOP		= 0;
  BOTTOM	= SIZE_Y;
  ORIGIN	= __ORIGIN = 0;

  STATE		= ESnormal;
  QUES		= 0;
  TRANSLATE	= set_translate(NORM_MAP);
  G0_CHARSET	= NORM_MAP;
  G1_CHARSET	= GRAF_MAP;
  CHARSET	= 0;
  NEED_WRAP	= 0;

/*	report_mouse	= 0;	*/

  MODE_FLAGS	= FL_AUTO_WRAP | FL_CURSOR_MODE;
  UTF_COUNT	= 0;

  set_kbd(VC_REPEAT);
  clr_kbd(VC_CKMODE);
  clr_kbd(VC_APPLIC);
  clr_kbd(VC_CRLF);

  vc_kbd[VC_NUM].kbdmode = VC_XLATE;
  vc_kbd[VC_NUM].lockstate = 0;
  vc_kbd[VC_NUM].ledmode = LED_SHOW_FLAGS;
  vc_kbd[VC_NUM].ledflagstate = vc_kbd[VC_NUM].default_ledflagstate;
  set_leds();

  TAB_STOP[0]	= 0x01010100;
  TAB_STOP[1]	= TAB_STOP[2] = TAB_STOP[3] = TAB_STOP[4] = 0x01010101;

  default_attr(cons);
  update_attr(cons); 
  gotoxy(cons,0,0);
  set_cursor(cons);
  set_pointer(cons);
  save_cur(cons);

  if (do_clear)
    csi_J(cons,2);

  if (SET_ORIGIN)
    SET_ORIGIN (ORIGIN);
}

/*\
|*| ---------------------------------------------------------------------------
|*|	The tty_driver functions.
|*| ---------------------------------------------------------------------------
\*/

static int
con_open(struct tty_struct *tty, struct file *filp)
{
  unsigned int index = MINOR(tty->device) - tty->driver.minor_start;
  struct kgi_vc *cons = vcons[index];

  tty->driver_data = NULL;

  if (!cons)			/* possible hardware driver will vmalloc  */
    return -ENODEV;		/* vcons[index] and then call us.         */
				/* Note that the caller has to initialize */
  tty->driver_data = cons;	/* vcons[index].			  */

  if (!tty->winsize.ws_row && !tty->winsize.ws_col)
    {
      tty->winsize.ws_row = vcons[index]->size.y;
      tty->winsize.ws_col = vcons[index]->size.x;
    }
  return 0;
}

static void
con_stop(struct tty_struct *tty)
{
  struct kgi_vc *cons = tty->driver_data;

  set_kbd_led(VC_SCROLLOCK);
  set_leds();
}

static void
con_start(struct tty_struct *tty)
{
  struct kgi_vc *cons = tty->driver_data;

  clr_kbd_led(VC_SCROLLOCK);
  set_leds();
}

static int
con_write_room(struct tty_struct *tty)
{
  return tty->stopped ? 0 : 4096;    /* No limit, really; we're not buffering */
}

static int
con_chars_in_buffer(struct tty_struct *tty)
{
  return 0;			     /* we are not buffering */
}

/* con_throttle and con_unthrottle are only used for paste_selection	*/
/* which has to stuff in a large number of characters...		*/

static void
con_throttle(struct tty_struct *tty)
{
}

static void
con_unthrottle(struct tty_struct *tty)
{
  struct kgi_vc *cons = tty->driver_data;

  wake_up_interruptible(&(cons->paste_wait));
}

static int
con_write(struct tty_struct *tty, int from_user, unsigned char *buf, int count)
{
  int c, tc, ok, n = 0;
  struct kgi_vc *cons = tty->driver_data;
  
/* !!! if (currcons == sel_cons) clear_selection(); */

  disable_bh(KEYBOARD_BH);
  while (!tty->stopped && count)
   {
     c = from_user ? get_fs_byte(buf) : *buf;
     buf++; n++; count--;

     if (UTF8_ENCODE)
       { /* Combine UTF-8 into Unicode */
         /* Incomplete characters silently ignored */
         if (c > 0x7f)
           {   
             if (UTF_COUNT > 0 && (c & 0xc0) == 0x80)
               {
                 UTF_CHAR = (UTF_CHAR << 6) | (c & 0x3f);
                 UTF_COUNT--;
                 if (UTF_COUNT == 0)
                   c = UTF_CHAR;
                  else
                   continue;
               }
              else
               {
                 if ((c & 0xe0) == 0xc0)
                   {
                     UTF_COUNT = 1;
                     UTF_CHAR  = (c & 0x1f);
                   }
                  else
                   if ((c & 0xf0) == 0xe0)
                     {
                       UTF_COUNT = 2;
                       UTF_CHAR = (c & 0x0f);
                     }
                    else
                     UTF_COUNT = 0;
                 continue;
               }
           }
          else
           UTF_COUNT = 0;

         /* Now try to find out how to display it */
         tc = conv_uni_to_pc(c);
         if (tc == -1 || tc == -2)
           continue;
         if (tc == -3 || tc == -4)	/* hashtable not valid */
           {				/* or symbol not found */
             tc = (c <= 0xff) ? TRANSLATE[c] : 040;
             ok = 0;
           }
          else
           ok = 1;
       }
      else	/* if (UTF8) ... */
       {
         tc = TRANSLATE[TOGGLE_META ? (c | 0x80) : c];
         ok = 0;
       }

     if (CURSOR.state & PF_FB_MODIFIED)
       if (CURSOR.hide)
         CURSOR.hide(&CURSOR, POS);

     if (POINTER.state & PF_FB_MODIFIED)
       if (POINTER.hide) 
         POINTER.hide(&POINTER, PPOS);

     /*\
     |*|	If the original code was < 32 we only allow a
     |*|	glyph to be displayed if the code is not normally
     |*|	used (such as for cursor movement) or if the
     |*|	disp_ctrl mode has been explicitly enabled.
     |*|	Note: ESC is *never* allowed to be displayed as
     |*|	that would disable all escape sequences!
     \*/

     if (   !ok && tc && (c >= 32 || (DISP_CTRL && c != 0x1b) 
         || !((CTRL_ACTION >> c) & 1)))
       ok = 1;

     if (STATE == ESnormal && ok)
       {
         if (NEED_WRAP) 
           {
             cr(cons);
             lf(cons);
           }
         if (INSERT_MODE)
           insert_char(cons);

	 FB[POS] = (ATTR << 8) + tc;

         if (X == (SIZE_X - 1))
           NEED_WRAP = !!AUTO_WRAP;
          else
           {
             X++;
             POS++;
           }
         continue;
       }

     /*\
     |*|	Control characters can be used in the _middle_
     |*|	of an escape sequence.
     \*/

     switch (c)
      {
	case ASCII_BEL:	kd_mksound(0x637, HZ/8);
			continue;
	
	case ASCII_BS:	bs(cons);
			continue;

	case ASCII_TAB:	POS -= X;
			while (X < SIZE_X - 1) 
			 {
			   X++;
			   if (TAB_STOP[X >> 5] & (1 << (X & 31)))
			     break;
			 }
			POS += X;
			continue;

	case ASCII_LF:
	case ASCII_VT:
	case ASCII_FF:	lf(cons);
			if (!is_kbd(VC_CRLF))
                          continue;

	case ASCII_CR:	cr(cons);
			continue;

	case ASCII_SO:	CHARSET = 1;
			TRANSLATE = set_translate(G1_CHARSET);
			  MODE_FLAGS |= FL_DISP_CTRL;
			continue;

 	case ASCII_SI:	CHARSET = 0;
			TRANSLATE = set_translate(G0_CHARSET);
			  MODE_FLAGS &= ~DISP_CTRL;
			continue;

	case ASCII_CAN:
	case ASCII_SUB:	STATE = ESnormal;
			continue;

	case ASCII_ESC:	STATE = ESesc;
			continue;

	case 127:	del(cons);
			continue;

	case 128+27:	STATE = ESsquare;
			continue;
      }

     switch(STATE)
      {
	case ESesc:	STATE = ESnormal;
			switch (c)
			 {
			   case '[':	STATE = ESsquare;
					continue;

			   case '%':	STATE = ESpercent;
					continue;

			   case 'E':	cr(cons);
					lf(cons);
					continue;

			   case 'M':	ri(cons);
					continue;

			   case 'D':	lf(cons);
					continue;

			   case 'H':	TAB_STOP[X >> 5] |= (1 << (X & 31));
					continue;

			   case 'Z':	respond_ID(tty);
					continue;

			   case '7':	save_cur(cons);
					continue;

			   case '8':	restore_cur(cons);
					continue;

			   case '(':	STATE = ESsetG0;
					continue;

			   case ')':	STATE = ESsetG1;
					continue;

			   case '#':	STATE = EShash;
					continue;

			   case 'c':	reset_terminal(cons,1);
					continue;

			   case '>':	/* Numeric keypad */
					clr_kbd(VC_APPLIC);
					continue;

			   case '=':	/* Appl. keypad */
					set_kbd(VC_APPLIC);
				 	continue;
			 }
			continue;

	case ESsquare:	for(NPAR = 0; NPAR < MAX_NPAR; NPAR++)
			  PAR[NPAR] = 0;
			NPAR = 0;
			STATE = ESgetpars;
			if (c == '[')
			  { /* Function key */
			    STATE=ESfunckey;
			    continue;
			  }
			if ((QUES = (c == '?')))
			  continue;

	case ESgetpars: if ((c == ';') && (NPAR < MAX_NPAR-1))
			  {
			    NPAR++;
			    continue;
			  }
			 else
			  if ((c >= '0') && (c <= '9'))
			    {
			      PAR[NPAR] *= 10;
			      PAR[NPAR] += c-'0';
			      continue;
			    }
			   else
			    STATE = ESgotpars;

	case ESgotpars:	STATE = ESnormal;
			switch(c)
			 {
			   case 'h':	set_mode(cons,1);
					continue;

			   case 'l':	set_mode(cons,0);
					continue;

			   case 'n':	if (!QUES)
					  if (PAR[0] == 5)
					    status_report(tty);
					   else
					    if (PAR[0] == 6)
					      cursor_report(cons,tty);
					continue;
			 }
			if (QUES)
			  {
			    QUES = 0;
			    continue;
			  }
			switch(c)
			 {
			   case 'G':
			   case '`':	if (PAR[0]) PAR[0]--;
					gotoxy(cons, PAR[0], Y);
					continue;

			   case 'A':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, X, Y - PAR[0]);
					continue;

			   case 'B':
			   case 'e':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, X, Y + PAR[0]);
					continue;

			   case 'C':
			   case 'a':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, X + PAR[0], Y);
					continue;

			   case 'D':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, X - PAR[0], Y);
					continue;

			   case 'E':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, 0, Y + PAR[0]);
					continue;

			   case 'F':	if (!PAR[0]) PAR[0]++;
					gotoxy(cons, 0, Y - PAR[0]);
					continue;

			   case 'd':	if (PAR[0]) PAR[0]--;
					gotoxy(cons, X, PAR[0]);
					continue;

			   case 'H':
			   case 'f':	if (PAR[0]) PAR[0]--;
					if (PAR[1]) PAR[1]--;
					gotoxy(cons, PAR[1], PAR[0]);
					continue;

			   case 'J':	csi_J(cons, PAR[0]);
					continue;

			   case 'K':	csi_K(cons, PAR[0]);
					continue;

			   case 'L':	csi_L(cons, PAR[0]);
					continue;

			   case 'M':	csi_M(cons, PAR[0]);
					continue;

			   case 'P':	csi_P(cons, PAR[0]);
					continue;

			   case 'c':	if (!PAR[0])
					  respond_ID(tty);
					continue;

			   case 'g':	if (!PAR[0])
					  TAB_STOP[X >> 5] &= ~(1 << (X & 31));
					 else
					  if (PAR[0] == 3)
					    {
					      TAB_STOP[0] = TAB_STOP[1] =
					      TAB_STOP[2] = TAB_STOP[3] =
					      TAB_STOP[4] = 0;
					    }
					continue;

			   case 'm':	csi_m(cons);
					continue;

			   case 'q':	/* DECLL - but only 3 leds */
					/* map 0,1,2,3 to 0,1,2,4 */
					if (PAR[0] < 4)
					  setledstate(vc_kbd + VC_NUM,
						      (PAR[0]<3) ? PAR[0] : 4);
					continue;

			   case 'r':	if (!PAR[0]) PAR[0]++;
					if (!PAR[1]) PAR[1] = SIZE_Y;
					/* Minimum allowed region is 2 lines */
					if (PAR[0] < PAR[1] &&
					    PAR[1] <= SIZE_Y)
					  {
					    TOP    = PAR[0]-1;
					    BOTTOM = PAR[1];
					    gotoxy(cons, 0, 0);
					  }
					continue;

			   case 's':	save_cur(cons);
					continue;

			   case 'u':	restore_cur(cons);
					continue;

			   case 'X':	csi_X(cons, PAR[0]);
					continue;

			   case '@':	csi_at(cons, PAR[0]);
					continue;

			   case ']':	setterm_command(cons);
					continue; /* setterm functions */
			 }
			continue;

	case ESpercent:	STATE = ESnormal;
			switch (c)
			 {
			   case '@':	/* defined in ISO 2022 */
					MODE_FLAGS &= ~FL_UTF8_ENCODE;
					continue;
			   case 'G':  /* prelim official escape code */
			   case '8':  /* retained for compatibility */
					MODE_FLAGS |= FL_UTF8_ENCODE;
					continue;
			 }
			continue;

	case ESfunckey:	STATE = ESnormal;
			continue;

	case EShash:	STATE = ESnormal;
			if (c == '8')
			  {
			    /* DEC screen alignment test. kludge :-) */
			    ERASE = (ERASE & 0xff00) | 'E';
			    csi_J(cons, 2);
			    ERASE = (ERASE & 0xff00) | ' ';
			  }
			continue;

	case ESsetG0:	if (c == '0') G0_CHARSET = GRAF_MAP; else 
			if (c == 'B') G0_CHARSET = NORM_MAP; else
			if (c == 'U') G0_CHARSET = NULL_MAP; else
			if (c == 'K') G0_CHARSET = USER_MAP;

			if (CHARSET == 0)
			  TRANSLATE = set_translate(G0_CHARSET);
			STATE = ESnormal;
			continue;

	case ESsetG1:	if (c == '0') G1_CHARSET = GRAF_MAP; else
			if (c == 'B') G1_CHARSET = NORM_MAP; else
			if (c == 'U') G1_CHARSET = NULL_MAP; else
			if (c == 'K') G1_CHARSET = USER_MAP;

			if (CHARSET == 1)
			  TRANSLATE = set_translate(G1_CHARSET);
			STATE = ESnormal;
			continue;

	default:	STATE = ESnormal;
      }
   }

  if (SYNC) (*SYNC)(cons);
  set_cursor(cons); 
  set_pointer(cons);

  FB_MODIFIED = 1;

  enable_bh(KEYBOARD_BH);
  return n;
}


#define VT_IS_IN_USE(i)	(console_driver.table[i] && console_driver.table[i]->count)
#define VT_BUSY(i)	(VT_IS_IN_USE(i) || i == fg_console)
/* !!! was additional  "|| i == sel_cons)" but we have no selection yet. */


/*\
|*|	console-specific ioctl's. We allow any console to be modified,
|*|	not just the fg_console.
\*/

static int
con_ioctl(struct tty_struct *tty, struct file * file,
          unsigned int cmd, unsigned long arg)
{
  struct kgi_vc *cons = tty->driver_data;
  struct kbd_struct *kbd = vc_kbd+VC_NUM;
  int i = 0, err;
  int perm = (current->tty == tty) || suser();

  switch (cmd)
   {
     case KIOCSOUND:

	if (!perm)
	  return -EPERM;
	kd_mksound((unsigned int)arg, 0);
	return 0;

     case KDMKTONE:

	if (!perm)
	  return -EPERM;
	{
	  unsigned int ticks = HZ * ((arg >> 16) & 0xFFFF) / 1000;

	  kd_mksound(arg & 0xffff, ticks);
	  if (ticks == 0)
	    kd_nosound(0);
	  return 0;
	}

     case KDGKBTYPE:
	/*\
	|*|	this is naive.
	\*/

	if (! (err = verify_area(VERIFY_WRITE, (void *)arg, sizeof(u_char))) )
 	  put_fs_byte(KB_101, (char *)arg);
	return err;

     case KDADDIO:	/* These are gone. Otherwise GGI would be useless. */
     case KDDELIO:

     case KDENABIO:
     case KDDISABIO:

     case KDMAPDISP:
     case KDUNMAPDISP:
	return -ENOIOCTLCMD;


     case KDSETMODE:	/* The new VC concept doesn't need this. Consoles */
     case KDGETMODE:	/* are just there. No process control allowed.	  */
	return -ENOIOCTLCMD;

     /*\
     |*|	Keyboard mode IOC's.
     \*/

     case KDSKBMODE:
	if (!perm)
	  return -EPERM;

	switch(arg)
	 {
	   case K_RAW:

		kbd->kbdmode = VC_RAW;
		break;

	   case K_MEDIUMRAW:

		kbd->kbdmode = VC_MEDIUMRAW;
		break;

	   case K_XLATE:

		kbd->kbdmode = VC_XLATE;
		compute_shiftstate();
		break;

	   case K_UNICODE:

		kbd->kbdmode = VC_UNICODE;
		compute_shiftstate();
		break;

	   default:
		return -EINVAL;
	 }

	if (tty->ldisc.flush_buffer)
	  tty->ldisc.flush_buffer(tty);

	return 0;

     case KDGKBMODE:

	err = verify_area(VERIFY_WRITE, (void *)arg, sizeof(unsigned long));
	if (!err)
	  {
	    unsigned long ucval = 0;
	    switch(kbd->kbdmode)
	     {
	       case VC_RAW:

		  ucval = K_RAW;
		  break;

	       case VC_MEDIUMRAW:

		  ucval = K_MEDIUMRAW;
		  break;

	       case VC_UNICODE:

		  ucval = K_UNICODE;
		  break;

	       case VC_XLATE:
		  ucval = K_XLATE;
		  break;

	       default:
		  err = -EINVAL;
	     }
	    if (!err) put_fs_long(ucval, (unsigned long *) arg);
	  }
	return err;


     /*\
     |*|	this could be folded into KDSKBMODE, but for compatibility
     |*|	reasons it is not so easy to fold KDGKBMETA into KDGKBMODE
     \*/
     case KDSKBMETA:

	switch(arg)
	 {
	   case K_METABIT:

	      clr_vc_kbd_mode(kbd, VC_META);
	      return 0;

	   case K_ESCPREFIX:

	      set_vc_kbd_mode(kbd, VC_META);
	      return 0;

	   default:
	      return -EINVAL;
	 }

     case KDGKBMETA:

	err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(unsigned long));
	if (!err)
	  {
	    unsigned long 
	    ucval = vc_kbd_mode(kbd, VC_META) ? K_ESCPREFIX : K_METABIT;
	    put_fs_long(ucval, (unsigned long *) arg);
	  }
	return err;

     case KDGETKEYCODE:
	{
	  struct kbkeycode * const a = (struct kbkeycode *)arg;
	  unsigned int sc;
	  int kc;

	  err = verify_area(VERIFY_WRITE, (void *)a, sizeof(struct kbkeycode));
	  if (err)
	    return err;

	  sc = get_fs_long((int *) &a->scancode);

	  if ((kc = getkeycode(sc)) < 0)
	    return kc;

	  put_fs_long(kc, (int *) &a->keycode);
	  return 0;
	}

     case KDSETKEYCODE:
	{
	  struct kbkeycode * const a = (struct kbkeycode *)arg;
	  unsigned int sc, kc;

	  if (!perm)
	    return -EPERM;

	  err = verify_area(VERIFY_READ, (void *)a, sizeof(struct kbkeycode));
	  if (err)
	    return err;

	  sc = get_fs_long((int *) &a->scancode);
	  kc = get_fs_long((int *) &a->keycode);
	  return setkeycode(sc, kc);
	}

     case KDGKBENT:
	{
	   struct kbentry * const a = (struct kbentry *)arg;
	   u_short *key_map, val;
	   u_char s;

	   err = verify_area(VERIFY_WRITE, (void *)a, sizeof(struct kbentry));
	   if (err)
	     return err;

	   if ((i = get_fs_byte((char *) &a->kb_index)) >= NR_KEYS)
	     return -EINVAL;
	   if ((s = get_fs_byte((char *) &a->kb_table)) >= MAX_NR_KEYMAPS)
	     return -EINVAL;

	   key_map = key_maps[s];
	   if (key_map)
             {
	       val = U(key_map[i]);
	       if ((kbd->kbdmode != VC_UNICODE) && (KTYP(val) >= NR_TYPES))
	         val = K_HOLE;
	     }
	    else
	     val = (i ? K_HOLE : K_NOSUCHMAP);
	   put_fs_word(val, (short *) &a->kb_value);
	   return 0;
	}

     case KDSKBENT:
	{
	  const struct kbentry * a = (struct kbentry *)arg;
	  u_short *key_map;
	  u_char s;
	  u_short v, ov;

	  if (!perm)
	    return -EPERM;
	  err = verify_area(VERIFY_READ, (void *)a, sizeof(struct kbentry));
	  if (err)
	    return err;

	  if ((i = get_fs_byte((char *) &a->kb_index)) >= NR_KEYS)
	    return -EINVAL;
	  if ((s = get_fs_byte((char *) &a->kb_table)) >= MAX_NR_KEYMAPS)
	    return -EINVAL;
	  v = get_fs_word(&a->kb_value);
	  if (!i && v == K_NOSUCHMAP)
            {
	       /* disallocate map */
               key_map = key_maps[s];
	       if (s && key_map)
                 {
	           key_maps[s] = 0;
	           if (key_map[0] == U(K_ALLOCATED))
	             {
	               kfree_s(key_map, sizeof(plain_map));
	               keymap_count--;
	             }
	         }
	       return 0;
	    }

	  if (KTYP(v) < NR_TYPES)
	    {
	      if (KVAL(v) > max_vals[KTYP(v)])
	        return -EINVAL;
	    }
	   else
	    if (kbd->kbdmode != VC_UNICODE)
	      return -EINVAL;

	  /* assignment to entry 0 only tests validity of args */
	  if (!i)
	    return 0;

	  if (!(key_map = key_maps[s]))
	    {
	      int j;

	      if (keymap_count >= MAX_NR_OF_USER_KEYMAPS && !suser())
	        return -EPERM;

	      key_map = (u_short *) kmalloc(sizeof(plain_map), GFP_KERNEL);
	      if (!key_map)
	        return -ENOMEM;
	      key_maps[s] = key_map;
	      key_map[0] = U(K_ALLOCATED);
	      for (j = 1; j < NR_KEYS; j++)
	        key_map[j] = U(K_HOLE);
	      keymap_count++;
	    }
	  ov = U(key_map[i]);
	  if (v == ov)
	    return 0;	/* nothing to do */

	  /* Only the Superuser can set or unset the Secure Attention Key. */
	  if (((ov == K_SAK) || (v == K_SAK)) && !suser())
	    return -EPERM;
	  key_map[i] = U(v);
	  if (!s && (KTYP(ov) == KT_SHIFT || KTYP(v) == KT_SHIFT))
	    compute_shiftstate();
	  return 0;
	}

     case KDGKBSENT:
	{
	  struct kbsentry *a = (struct kbsentry *)arg;
	  char *p;
	  u_char *q;
	  int sz;

	  err = verify_area(VERIFY_WRITE, (void *)a, sizeof(struct kbsentry));
	  if (err)
	  return err;
	  if ((i = get_fs_byte(&a->kb_func)) >= MAX_NR_FUNC || i < 0)
	    return -EINVAL;
	  sz = sizeof(a->kb_string) - 1;	/* sz should have been	*/
	  q = a->kb_string;			/* a struct member	*/
	  p = func_table[i];
	  if (p)
	    for( ; *p && sz; p++, sz--)
	      put_fs_byte(*p, q++);
	  put_fs_byte(0, q);
	  return ((p && *p) ? -EOVERFLOW : 0);
	}

     case KDSKBSENT:
	{
	  struct kbsentry * const a = (struct kbsentry *)arg;
	  int delta;
	  char *first_free, *fj, *fnw;
	  int j, k, sz;
	  u_char *p;
	  char *q;

	  if (!perm)
	    return -EPERM;
	  err = verify_area(VERIFY_READ, (void *)a, sizeof(struct kbsentry));
	  if (err)
	    return err;

	  if ((i = get_fs_byte(&a->kb_func)) >= MAX_NR_FUNC)
	    return -EINVAL;
	  q = func_table[i];

	  first_free = funcbufptr + (funcbufsize - funcbufleft);
	  for (j = i+1; j < MAX_NR_FUNC && !func_table[j]; j++) ;
	  if (j < MAX_NR_FUNC)
	    fj = func_table[j];
	   else
	    fj = first_free;

	  delta = (q ? -strlen(q) : 1);
	  sz = sizeof(a->kb_string); 	/* sz should have been
					   a struct member */
	  for (p = a->kb_string; get_fs_byte(p) && sz; p++,sz--)
	    delta++;
	  if (!sz)
	    return -EOVERFLOW;
	  if (delta <= funcbufleft) 	/* it fits in current buf */
	    {
	      if (j < MAX_NR_FUNC)
                {
	          memmove(fj + delta, fj, first_free - fj);
	          for (k = j; k < MAX_NR_FUNC; k++)
	            if (func_table[k])
	              func_table[k] += delta;
	        }
	      if (!q)
	        func_table[i] = fj;
	      funcbufleft -= delta;
	    }
	   else
	    {			/* allocate a larger buffer */
	      sz = 256;
	      while (sz < funcbufsize - funcbufleft + delta)
	        sz <<= 1;
	      fnw = (char *) kmalloc(sz, GFP_KERNEL);
	      if (!fnw)
	        return -ENOMEM;

	      if (!q)
	        func_table[i] = fj;
	      if (fj > funcbufptr)
	        memmove(fnw, funcbufptr, fj - funcbufptr);
	      for (k = 0; k < j; k++)
	       if (func_table[k])
	         func_table[k] = fnw + (func_table[k] - funcbufptr);

	      if (first_free > fj)
	        {
	          memmove(fnw + (fj - funcbufptr) + delta, fj, first_free - fj);
	          for (k = j; k < MAX_NR_FUNC; k++)
	            if (func_table[k])
	              func_table[k] = fnw + (func_table[k]-funcbufptr) + delta;
	        }
	      if (funcbufptr != func_buf)
	        kfree_s(funcbufptr, funcbufsize);
	      funcbufptr = fnw;
	      funcbufleft = funcbufleft - delta + sz - funcbufsize;
	      funcbufsize = sz;
	    }
	  for (p = a->kb_string, q = func_table[i]; ; p++, q++)
	    if (!(*q = get_fs_byte(p)))
	      break;
	  return 0;
	}

     case KDGKBDIACR:
	{
	  struct kbdiacrs *a = (struct kbdiacrs *)arg;

	  err = verify_area(VERIFY_WRITE, (void *) a, sizeof(struct kbdiacrs));
	  if (err)
	    return err;
	  put_fs_long(accent_table_size, &a->kb_cnt);
	  memcpy_tofs(a->kbdiacr, accent_table,
	  accent_table_size*sizeof(struct kbdiacr));
	  return 0;
	}

     case KDSKBDIACR:
	{
	  struct kbdiacrs *a = (struct kbdiacrs *)arg;
	  unsigned int ct;

	  if (!perm)
	    return -EPERM;
	  err = verify_area(VERIFY_READ, (void *) a, sizeof(struct kbdiacrs));
	  if (err)
	    return err;
	  ct = get_fs_long(&a->kb_cnt);
	  if (ct >= MAX_DIACR)
	    return -EINVAL;
	  accent_table_size = ct;
	  memcpy_fromfs(accent_table, a->kbdiacr, ct*sizeof(struct kbdiacr));
	  return 0;
	}

	/* the ioctls below read/set the flags usually shown in the leds */
	/* don't use them - they will go away without warning */
     case KDGKBLED:
	err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(unsigned char));
	if (err)
	  return err;

	put_fs_byte(kbd->ledflagstate | (kbd->default_ledflagstate << 4),
		    (char *) arg);
	return 0;

     case KDSKBLED:
	if (!perm)
	  return -EPERM;
	if (arg & ~0x77)
	  return -EINVAL;
	kbd->ledflagstate = (arg & 7);
	kbd->default_ledflagstate = ((arg >> 4) & 7);
	set_leds();
	return 0;

	/* the ioctls below only set the lights, not the functions */
	/* for those, see KDGKBLED and KDSKBLED above */
     case KDGETLED:
	err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(unsigned char));
	if (err)
	  return err;
	put_fs_byte(getledstate(), (char *) arg);
	return 0;

     case KDSETLED:
	if (!perm)
	  return -EPERM;
	setledstate(kbd, arg);
	return 0;

     case KDSIGACCEPT:
	/*\
	|*|	A process can indicate its willingness to accept signals
	|*|	generated by pressing an appropriate key combination.
	|*|	Thus, one can have a daemon that e.g. spawns a new console
	|*|	upon a keypress and then changes to it.
	|*|	Probably init should be changed to do this (and have a
	|*|	field ks (`keyboard signal') in inittab describing the
	|*|	desired action), so that the number of background daemons
	|*|	does not increase.
	\*/
	{
	  extern int spawnpid, spawnsig;
	  if (!perm)
	    return -EPERM;
	  if (arg < 1 || arg > NSIG || arg == SIGKILL)
	    return -EINVAL;
	  spawnpid = current->pid;
	  spawnsig = arg;
	  return 0;
	}

     case PIO_SCRNMAP:
	if (!perm)
	  return -EPERM;
	return con_set_trans((char *)arg);

     case GIO_SCRNMAP:
	return con_get_trans((char *)arg);

     case PIO_UNIMAPCLR:
	{
	  struct unimapinit ui;
	  if (!perm)
	    return -EPERM;
	  err = verify_area(VERIFY_READ, (void *)arg, sizeof(struct unimapinit));
	  if (err)
	    return err;
	  memcpy_fromfs(&ui, (void *)arg, sizeof(struct unimapinit));
	  con_clear_unimap(&ui);
	  return 0;
	}

     case PIO_UNIMAP:
	{
	  struct unimapdesc *ud;
	  u_short ct;
	  struct unipair *list;

	  if (!perm)
	    return -EPERM;

	  err = verify_area(VERIFY_READ, (void *)arg, sizeof(struct unimapdesc));
	  if (err)
	    return err;

	  ud = (struct unimapdesc *) arg;
	  ct = get_fs_word(&ud->entry_ct);
	  list = (struct unipair *) get_fs_long(&ud->entries);
	  err = verify_area(VERIFY_READ, (void *) list,
	                    ct*sizeof(struct unipair));
	  if (err)
	    return err;
	  return con_set_unimap(ct, list);
	}

     case GIO_UNIMAP:
 	{
	  struct unimapdesc *ud;
	  u_short ct;
	  struct unipair *list;

	  err = verify_area(VERIFY_WRITE, (void *)arg, sizeof(struct unimapdesc));
	  if (err)
	    return err;
	
	  ud = (struct unimapdesc *) arg;
	  ct = get_fs_word(&ud->entry_ct);
	  list = (struct unipair *) get_fs_long(&ud->entries);
	  if (ct)
	    err = verify_area(VERIFY_WRITE, (void *) list,
	                      ct*sizeof(struct unipair));
	  if (err)
	    return err;
	  return con_get_unimap(ct, &(ud->entry_ct), list);
	}
 
     default:
	return -ENOIOCTLCMD;
   }
}

/*\
|*| --------------------------------------------------------------------------
|*|	kernel interface & low level routines
|*| --------------------------------------------------------------------------
\*/

static int CRT_base = 0;

static void
set_origin(unsigned int offset)
{
  unsigned long flags;

  save_flags(flags); cli();
  outb_p(12, CRT_base); outb_p(offset >> 8, CRT_base + 1);
  outb_p(13, CRT_base); outb_p(offset, CRT_base + 1);
  restore_flags(flags);
}

static void
show_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  unsigned long flags;

  save_flags(flags); cli();
  cursor->state |= PF_SHOWN;
  outb_p(14, CRT_base); outb_p(offset >> 8, CRT_base + 1);
  outb_p(15, CRT_base); outb_p(offset, CRT_base + 1);
  restore_flags(flags);
}

static void
hide_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  unsigned long flags;

  if (!(cursor->state & PF_SHOWN))
    return;

  save_flags(flags); cli();
  cursor->state &= ~PF_SHOWN;
  outb_p(14, CRT_base); outb_p(0xFF, CRT_base + 1);
  outb_p(15, CRT_base); outb_p(0xFF, CRT_base + 1);
  restore_flags(flags);
}

void
scrollback(unsigned int lines)
{
   struct kgi_vc *cons = vcons[fg_console];

   if (lines > FB_SIZE_Y)
     lines = FB_SIZE_Y;
   lines = lines ? lines*FB_SIZE_X : (SIZE_Y/2)*FB_SIZE_X;
   __ORIGIN -= lines;
   if (__ORIGIN < 0)
     __ORIGIN = 0;
   if (SET_ORIGIN)
     SET_ORIGIN(__ORIGIN);
}

void
scrollfront(unsigned int lines)
{
  struct kgi_vc *cons = vcons[fg_console];
  
  if (lines > FB_SIZE_Y)
    lines = FB_SIZE_Y;
  lines = lines ? lines*FB_SIZE_X : (SIZE_Y/2)*FB_SIZE_X;
  __ORIGIN += lines;
  if (__ORIGIN > ORIGIN)
    __ORIGIN = ORIGIN;
  if (SET_ORIGIN)
    SET_ORIGIN(__ORIGIN);
}

static int printable = 0;

void
console_print(const char *b)
{
  struct kgi_vc *cons = vcons[fg_console < MAX_NR_CONSOLES 
                                 ? fg_console : fg_console - MAX_NR_CONSOLES];
  u_char c;
  static int printing = 0;

  if (!printable || printing || !cons || !FB)
    return;

  printing = 1;

    if (CURSOR.state & PF_FB_MODIFIED)
      if (CURSOR.hide)
        CURSOR.hide(&CURSOR, POS);

    if (POINTER.state & PF_FB_MODIFIED)
      if (POINTER.hide)
        POINTER.hide(&POINTER, PPOS);

    while ((c = *(b++)))
     {
       if ((c == ASCII_LF) || (c == ASCII_CR) || NEED_WRAP)
         {
           if (c != ASCII_CR)
             lf(cons);
           cr(cons);
           if ((c == ASCII_LF) || (c == ASCII_CR))
             continue;
         }
       FB[POS] = (ATTR << 8) + c;
       if (X == (SIZE_X-1))
         {
           NEED_WRAP = 1;
           continue;
         }
       X++; POS++;
     }

    if (SYNC) (*SYNC)(cons);
    set_cursor(cons); 
    set_pointer(cons);
    FB_MODIFIED = 1;

  printing = 0;
}

static int driver_changed;
static void (*setup_hardware)(int);

int
change_console_tty_ops(struct tty_driver *new_driver, void (*setup)(int))
{
  if (!driver_changed && new_driver)
    {
      void *tmp;
      #define XCHG(F)  if (new_driver->F)\
                         {\
                           tmp = console_driver.F;\
                           console_driver.F = new_driver->F;\
                           new_driver->F = tmp;\
                         }
      driver_changed = 1;
      XCHG(open);
      XCHG(close);
      XCHG(write);
      XCHG(write_room);
      XCHG(chars_in_buffer);
      XCHG(ioctl);
      XCHG(stop);
      XCHG(start);
      XCHG(throttle);
      XCHG(unthrottle); 
      #undef XCHG
      setup_hardware = setup;
      new_driver->minor_start = console_driver.minor_start;
      new_driver->table       = console_driver.table;
      return 0;
    }
  return 1;
}

void
reset_console_tty_ops(void)
{
  driver_changed = 0;
  console_driver.open		= con_open;
  /* console_driver.close	= con_close; */
  console_driver.write		= con_write;
  console_driver.write_room	= con_write_room;
  console_driver.chars_in_buffer= con_chars_in_buffer;
  console_driver.ioctl		= con_ioctl;
  console_driver.stop		= con_stop;
  console_driver.start		= con_start;
  console_driver.throttle	= con_throttle;
  console_driver.unthrottle	= con_unthrottle;

  setup_hardware = NULL;	/* well, we don't how to do that (here). */ 
}

void 
change_console(int want)
{
  if (want < MAX_NR_CONSOLES)	/* text screen ? */
    {
      if (vcons[want])				/* console valid ?	  */
        {
          if (setup_hardware)
            (*setup_hardware)(want);
  	  fg_console = want;
        }
    }
   else				/* graphics screen ? */
    {
      if (ev_buffer[want-MAX_NR_CONSOLES])	/* graphics screen valid ? */
        {
          if (setup_hardware)
            (*setup_hardware)(want);
	  fg_console = want;
        }
    }
}

long
con_init(long kmem_start)
{
  /*\
  |*|	Initialize console during boot. We just determine the hardware and set
  |*|	up a simple virtual console with scrollback as vcons[0] and 
  |*|	fg_console = 0.
  \*/
  struct kgi_vc *cons;
  char *display_desc;
  unsigned int fb_memsize;

  memset(vcons, 0, sizeof(struct kgi_vc *)*MAX_NR_CONSOLES);
  memset(&console_driver, 0, sizeof(console_driver));

  console_driver.magic		= TTY_DRIVER_MAGIC;
  console_driver.name		= "tty";
  console_driver.name_base	= 1;
  console_driver.major		= TTY_MAJOR;
  console_driver.minor_start	= 1;
  console_driver.num		= MAX_NR_CONSOLES;
  console_driver.type		= TTY_DRIVER_TYPE_CONSOLE;
  console_driver.init_termios	= tty_std_termios;
  console_driver.flags		= TTY_DRIVER_REAL_RAW;
  console_driver.refcount	= &console_refcount;
  console_driver.table		= console_table;
  console_driver.termios	= console_termios;
  console_driver.termios_locked	= console_termios_locked;

  reset_console_tty_ops();

  if (tty_register_driver(&console_driver))
    panic("Couldn't register console driver\n");

#if 0
  timer_table[BLANK_TIMER].fn = blank_screen;
  timer_table[BLANK_TIMER].expires = 0;
  if (blankinterval)
    {
      timer_table[BLANK_TIMER].expires = jiffies+blankinterval;
      timer_active |= 1<<BLANK_TIMER;
    }
#endif

  last_console = fg_console = 0;
  boot_console = vcons[0] = cons = (struct kgi_vc *) kmem_start;
  kmem_start += sizeof(struct kgi_vc);
  memset(cons, 0, sizeof(struct kgi_vc));

  SIZE_X = ORIG_VIDEO_COLS;	/* set logical size */
  SIZE_Y = ORIG_VIDEO_LINES;

  CURSOR.hide = hide_cursor;
  CURSOR.show = show_cursor;
  SET_ORIGIN  = set_origin;

  if (ORIG_VIDEO_MODE == 7)	/* determine framebuffer size and location */
    {				/* we have a monochrome display */
      FB = (u_short *) 0x000B0000;
      CRT_base = 0x3B4;
      if ((ORIG_VIDEO_EGA_BX & 0xff) != 0x10)	/* EGA with Mono monitor ? */
        {
          fb_memsize = 0x000B8000 - 0x000B0000;
          display_desc = "EGA+";
        }
       else					/* MDA adapter             */
        {
          fb_memsize = 0x000B2000 - 0x000B0000;
          display_desc = "*MDA";
        }
    }
   else				/* ok, it is color. */
    {
      CAN_DO_COLOR = 1;
      FB = (u_short *) 0x000B8000;
      CRT_base = 0x3D4;
      if ((ORIG_VIDEO_EGA_BX & 0xff) != 0x10)
        {					/* either EGA/VGA	  */
          fb_memsize = 0x000C0000 - 0x000B8000;
          display_desc = "EGA+";
        }
       else
        {					/* or CGA		  */
          fb_memsize = 0x000BA000 - 0x000B8000;
          display_desc = "*CGA";
        }
    }

  FB_SIZE_X = SIZE_X;
  FB_SIZE_Y = fb_memsize/(sizeof(u_short)*FB_SIZE_X);
  FB_END    = FB_SIZE_X*FB_SIZE_Y;

  ORIGIN    = 0;
  SCR_END   = SIZE_Y*FB_SIZE_X;

  DEF_COLOR = 0x07;
  ULCOLOR   = 0x0F;
  HALFCOLOR = 0x08;

  reset_terminal(cons, 0);
  gotoxy(cons, ORIG_X, ORIG_Y);
  csi_J(cons, 0);

  printable = 1;
  printk("%s %s console, %hix%hi (%hix%hi), max %i console(s),"
         " kgi_vc is %i bytes\n", 
         CAN_DO_COLOR ? "color" : "mono", display_desc,
         SIZE_X, SIZE_Y, FB_SIZE_X, FB_SIZE_Y, MAX_NR_CONSOLES,
         sizeof(struct kgi_vc));

  register_console(console_print);
  return kmem_start;
}
