/*\
|*| ---------------------------------------------------------------------------
|*|	S3 SDAC/Trio32/Trio64 clock chip driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger [seeger@physik.tu-chemnitz.de].
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*| ---------------------------------------------------------------------------
|*|	This code was derived from the following source of information:
|*|	[1]	80C716 SDAC Data Book (c) 1994 by S3 Incorporated
|*|	[2]	Trio32/Trio64 Data Book (c) 1994 by S3 Incorporated
\*/

#define	MAINTAINER	"Steffen Seeger, seeger@physik.tu-chemnitz.de"

#include <linux/kernel.h>

#include "graphdev.h"
#include "kgi.h"

#ifdef TRIO
  #include "S3/Trio.h"

  static struct kgi_clock clock =
  {
	"S3 Incorporated",	/* manufacturer		*/
	"Trio32/Trio64",	/* model		*/
	GM_ALL,			/* modes		*/ 
	0,			/* flags		*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/

	CLK_TYPE_PROG,		/* clock type		*/
	{{ 25000000, 110000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
  };
#else
  #include "S3/sdac.h"

  static struct kgi_clock clock =
  {
	"S3 Incorporated",	/* manufacturer		*/ 
	"80C716-ME/MG (SDAC)",	/* model		*/
	GM_ALL,			/* supported modes 	*/
	0,			/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/

	CLK_TYPE_PROG,		/* clock type		*/
	{{   400000, 119000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
  };
#endif

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

/*\
|*|	Clock driver dependend #define's
\*/

#define CLK_VERSION "0.00 ALPHA"

#define	fref	14318180   /* Hz, must be within 14..15 MHz */ 

#define F(M,N,FVCO)  (((M)+2)*(FVCO)/((N)+2))

#define	INVALIDCLK 0xFFFF

static unsigned short clk_last_cmd = 0xFFFF;
static int 	      clk_last_freq = -1;

void
setPLL(unsigned short cmd, int clk)
{
  #ifdef TRIO
    #warning implement setPLL !!!
  #else
    outbDAC(clk ,DAC_PLL_W_INDEX);	/* select clock generator	*/
    outbDAC(cmd & 0xFF, DAC_PLL_DATA);	/* 1st byte			*/
    outbDAC(cmd >> 8, DAC_PLL_DATA);	/* 2nd byte			*/
  #endif
}

unsigned short
calcfreq(int *freq, enum kgi_tm_cmd cmd)
{
  int fvco = *freq;
  int hf   = *freq;
  int div  = 0;
  int n;
  struct { unsigned char m; unsigned char n; int delta; }
         best = { 0xFF, 0xFF, 0x7FFFFFFF };    
 
  if(*freq == clk_last_freq)	/* we never get called with freq == -1 */
    return clk_last_cmd;
  
  while((div < 4) && (hf < 135000000))
    { div += 1; hf += hf; }
 
  if((hf < 135000000) || (270000000 < hf) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz)\n",__FILE__,__LINE__,\
                   *freq));
      return INVALIDCLK;
    }

  fvco  = fref >> div;

  for(n = 1; n < 32; n--)
    { 
      register int m = hf*(n+2)/fref - 2;
      register int delta = *freq - F(m,n,fvco);
 
      delta = (cmd == CMD_LOWER) ? delta : -delta;

      if(m > 127) break;
      if((delta < 0) || (delta > best.delta)) continue;

      best.n = n; best.m = m; best.delta = delta;
    }

  if((best.n > 31) || (best.m > 127) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz)\n",__FILE__,__LINE__,\
                   *freq)); 
      return INVALIDCLK; 
    }
   else
    {
      TRACE(printk("%s: request for DCLK = %i Hz,", __FILE__,*freq ));
      *freq = clk_last_freq = F(best.m,best.n,fref >> div);
      TRACE(printk(" did %i Hz.\n", *freq));
      return clk_last_cmd = best.m | (best.n << 8) | (div << 13);
    }
}

/*\
|*|     GGI clock driver interface
\*/

static unsigned short PLL_save[4];
static unsigned char clk, PLL_ctrl;

int
kgi_clock_init(void)
{
  /*\
  |*|	Detection is done by reading CLK0 and CLK1 frequencies. As these are
  |*|	read only it should be save enough. (For the SDAC, the Trio will be 
  |*|	detected by the chipset driver.)
  \*/

  #ifdef TRIO
    #warning implement kgi_clock_init !!!	
  #else
    clk = getCLK();
    outbDAC(0, DAC_PLL_R_INDEX);		/* save PLL registers 0 - 15 */
    insbDAC(DAC_PLL_R_INDEX, &PLL_save, 8);
    /*
    if((PLL_save[0] != ...) || (PLL_save[1] != ...))
      return FALSE;
    */
    outbDAC(DAC_PLL_CONTROL, DAC_PLL_R_INDEX);
    PLL_ctrl = inbDAC(DAC_PLL_DATA);
  #endif
  
  printk("%s %s driver v"CLK_VERSION".\n", clock.manufact, clock.model);
  printk(
"NOTE: This driver has not been tested yet. Please report successful testing or
any problems that encounter to " MAINTAINER "
Please send also the following numbers: CLK0 = %.4X, CLK1 = %.4X
Thanks for brave alpha testing! ;-)\n", PLL_save[0], PLL_save[1]);
  return EOK;
}

void
kgi_clock_done(void)
{
  #ifdef TRIO
    #warning implement kgi_clock_done !!!
  #else
    setCLK(clk);
    setPLL(PLL_save[3], 3);
    outbDAC(DAC_PLL_CONTROL, DAC_PLL_W_INDEX);  
    outbDAC(PLL_ctrl, DAC_PLL_DATA);
  #endif
  TRACE(printk("%s %s clock driver removed.\n", clock.manufact, clock.model));
}

struct kgi_clock
*kgi_clock_get_info (void)
{ 
  return &clock;
}

int
kgi_clock_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newfreq = (cmd == CMD_PROPOSE) 
                    ? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->clock = INVALIDCLK;
	return EOK;

     case CMD_LOWER:
	if (newfreq < MIN_FREQ)
	  {
	    TRACE(printk("%s: requested DCLK too low (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -ECLOCK;
	  }

   	if (newfreq > MAX_FREQ)	/* meet limits */
	  newfreq = MAX_FREQ;
	
	mode->clock = calcfreq(&newfreq, cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;
			
     case CMD_RAISE:
	if (newfreq < MAX_FREQ)
	  {
	    TRACE(printk("%s: requested DCLK too high (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -ECLOCK;
	  }

	if (newfreq < MIN_FREQ)
	  newfreq = MIN_FREQ;

	mode->clock = calcfreq(&newfreq, cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -ECLOCK : EOK;

     case CMD_CHECK:
	if ((newfreq < MIN_FREQ) || (newfreq > MAX_FREQ))
	  return -ECLOCK;

	mode->clock = calcfreq(&newfreq, CMD_LOWER);

	if (   (mode->clock != INVALIDCLK)
	    && (   ((mode->dclk*mode->clk.mul/mode->clk.div) - newfreq) 
		 < ((mode->dclk*mode->clk.mul/mode->clk.div)*1/100)))
	  return EOK;

	return -ECLOCK;

     default:
	INTERNALERROR;
	return -ECLOCK;
    }
}

void
kgi_clock_set_mode(struct kgi_mode *mode)
{
  if(mode->clock != INVALIDCLK)
    {
      #ifdef TRIO
	#warning implement kgi_clock_set_mode !!
      #else
        /* always use CLK3, which is reprogrammed according to the request. */
        setPLL(mode->clock, 3);

	outbDAC(DAC_PLL_CONTROL, DAC_PLL_W_INDEX);
	outbDAC(0, DAC_PLL_DATA);

	setCLK(3);
      #endif
    }
   else
    { /* This should never happen !! */
      INTERNALERROR;
    }
}

int
kgi_clock_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	       unsigned long arg)
{
  return -ECLOCK;
}
