/*\
|*| --------------------------------------------------------------------------
|*|     KGI kernel driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by 
|*|
|*|	  Andreas Beck,		[becka@sunserver1.rz.uni-duesseldorf.de]
|*|	  Steffen Seeger	[seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#undef DEBUG


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include "graphdev.h"
#include <linux/tty.h>
#include <linux/kbd.h>
#include <linux/console.h>

#define	VC_VFB_SIZE	(3*4096-sizeof(struct kgi_vc))	/* size of virtual fb */

char kernel_version[] = UTS_RELEASE;

unsigned long fb_graphics;	/* phys. address of the graphics framebuffer  */
size_t	      fb_graph_len;	/* size of the graphics framebuffer           */
unsigned long fb_text;		/* physical address of the text framebuffer   */
size_t	      fb_text_len;	/* size of the text framebuffer		      */

static struct kgi_mode textmode[MAX_NR_CONSOLES];
static struct kgi_state *graph[MAX_NR_CONSOLES];

#ifdef DEBUG

static void
printk_mode (struct kgi_mode *mode, char *msg)
{
  #define M(X) mode->X
  #define R(X) mode->request.X
  #define T(X) M(X.width), M(X.blankstart), M(X.syncstart), M(X.syncend),\
		M(X.blankend), M(X.total), M(X.polarity)
  printk(msg);
  printk("mode = {\n  {{ %i, %i},  /* visible */\n",
			R(visible.x),R(visible.y));
  printk("   { %i, %i},   /* virtual */\n",R(virtual.x),R(virtual.y));
  printk("   { %i, %i},   /* size [mm] */\n",R(size.x),R(size.y));
  printk("   %s,\n", GRAPHMASK(R(graphtype)));
  printk("   { %i, %i}},   /* textgrid */\n", R(textgrid.x),R(textgrid.y));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n", T(x));
  printk("   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n", T(y));
  printk("   { %i, %i},   /* magnify */\n", M(magnify.x), M(magnify.y));
  printk("   %x,          /* flags   */\n", M(flags));
  printk("   %i,          /* dclk    */\n", M(dclk));
  printk("   { %i, %i},   /* clk     */\n", M(clk.mul), M(clk.div));
  printk("   { %i, %i}    /* lclk    */\n}\n", M(lclk.mul), M(lclk.div));

  #undef M
  #undef R
  #undef T
}
#endif /* #ifdef DEBUG */

#define	MAX_EVENTS	16	/* *must* be a power of 2 !!	*/

struct kgi_state			/* struct to keep state */
{
   int			graph_num;	/* index in graph[]	*/
				/* hardware state	*/
   struct kgi_mode	mode;		/* video mode		*/
   ggi_font		font0, font1;	/* font info (textmode)	*/
   ggi_palette		palette;	/* palette		*/
   ggi_color		color[256];	/* palette entries	*/
   ggi_coord		origin;		/* pos of visible frame */

				/* event reporting	*/
   event_buffer		ev_buffer;	/* event buffer status	*/
   ggi_event		event[MAX_EVENTS]; /* event buffer entries */
};

static void
kgi_free_state(struct kgi_state *state)
{
  if (state)
    {
       if (state->font0.free_data)
         vfree(state->font0.data);

       if (state->font1.free_data)
         vfree(state->font1.data);

       vfree(state);
    }
}

static int
kgi_check_mode(struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int cnt = 10;

  if (cmd == CMD_PROPOSE)
    {
      memset(&(mode->x), 0, sizeof(struct kgi_mode) - sizeof(ggi_mode));
      switch(mode->request.graphtype)
       {
	 case GT_TEXT:			/* propose a standard text mode */
		mode->palette = &default_text_palette;
		mode->font0   = &default_font;
		mode->font1   = &default_font;

		if (   ((mode->request.virtual.x/mode->request.textgrid.x) 
		      *(mode->request.virtual.y/mode->request.textgrid.y)
		      *sizeof(u_short))
		    > VC_VFB_SIZE)
		  return -ENOMEM;
		break;
	 case GT_4BIT:			/* indexed colors, but no text font */
	 case GT_8BIT:
		mode->palette  = &default_graphics_palette;
		mode->font0    = mode->font1 = NULL;
		break;
	 case GT_15BIT:			/* only gamma correction makes sense */
	 case GT_16BIT:			
	 case GT_24BIT:
	 case GT_32BIT:
		mode->palette  = &default_gamma_map;
		mode->font0    = mode->font1 = NULL;
	        break;
	 default:
		INTERNALERROR;
       }
    }

  while (cnt-- && !(cmd == CMD_READY))
   {
     register int err;
     TRACE(printk("check_mode: cmd = %i; cnt = %i\n", cmd, cnt));
     if ((err = kgi_clock_check_mode  (mode,  cmd)))  return err;
     TRACE(printk("check_mode: clock ok.\n"));	 
     if ((err = kgi_ramdac_check_mode (mode,  cmd)))  return err;
     TRACE(printk("check_mode: ramdac ok.\n"));
     if ((err = kgi_chipset_check_mode(mode,  cmd)))  return err;
     TRACE(printk("check_mode: chipset ok.\n"));
     if ((err = kgi_monitor_check_mode(mode, &cmd)))  return err;
     TRACE(printk("check_mode: monitor ok."));

     TRACE(printk_mode(mode, "mode:\n"));
   }

  TRACE(printk_mode(mode, "mode ok:\n"));
 
  return (++cnt) ? EOK : -EDRIVER;
}

static void
kgi_set_mode(struct kgi_mode *mode)
{
  static ggi_mode currmode;

  if (memcmp(&currmode, &(mode->request), sizeof(currmode)))
    {					/* avoid some stress on hardware by  */
      kgi_chipset_set_mode (mode);	/* unneccessary mode initialisations */
      currmode = mode->request;
    }
}

/*\
|*|	virtual console handling
\*/

static void
vc_reset(struct kgi_vc *cons, ggi_mode *mode)
{
  SIZE_X	= mode->visible.x / mode->textgrid.x;
  SIZE_Y	= mode->visible.y / mode->textgrid.y;

  ORIGIN	= 0;
  __ORIGIN	= 0;
  SCR_END	= SIZE_Y*FB_SIZE_X;

  DEF_COLOR	= 0x07;
  ULCOLOR	= 0x0F;
  HALFCOLOR	= 0x08;

  reset_terminal(cons, 0);
}

static void
vc_reset_to_mode(int vc, ggi_mode *mode)
{
  struct kgi_vc *cons = vcons[vc];
  unsigned int size   = VMALLOCED;
  unsigned short *d;

  memset(cons, 0, sizeof(struct kgi_vc));

  memset(textmode + vc, 0, sizeof(textmode[vc]));
  textmode[vc].request = *mode;
  kgi_check_mode(textmode + vc, CMD_PROPOSE);

  if (fg_console == vc)			/* current console to be reset? */
    {
      kgi_set_mode(textmode + vc);	/* We have to set the mode, 	*/
      SET_ORIGIN  = textmode->set_origin;	/* the hardware support	*/
      CURSOR.hide = textmode->cursor_hide;	/* routines and also to	*/
      CURSOR.show = textmode->cursor_show;
      FB = (void *) fb_text;		/* map the hardware framebuffer	*/
    }
   else
    {
       /* !!! cursor and scroll back routines for 'switched back' mode */
       FB = (void*)(cons + 1);		/* this VC is in background mode*/
    }
   
  VMALLOCED = size;
  VC_NUM    = vc;
  CAN_DO_COLOR = (textmode[vc].flags & FL_CHIP_COLORTEXT) == FL_CHIP_COLORTEXT;
 
  FB_SIZE_X = textmode[vc].request.virtual.x / textmode[vc].request.textgrid.x;
  FB_SIZE_Y = textmode[vc].request.virtual.y / textmode[vc].request.textgrid.y;
  FB_END    = FB_SIZE_X*FB_SIZE_Y;

  vc_reset(cons, &(textmode[vc].request));

  size = FB_END; d = FB;
  while (size--)
    *(d++) = ERASE;
}

static int  
vc_allocate(int vc)
{
  /*\
  |*|	We waste some memory here because we allocate the same amount of
  |*|	memory for all VC's. Anyhow, what is left could be used for scroll-
  |*|	back when switched away. The (free) space after the VC state is
  |*|	used for the virtual framebuffer when in background.
  \*/

  size_t size = 3*4096;	/* number of bytes for VC state & framebuffer */

  /* !!! some locking to prevent ordinary users from taking too much memory */
  /* !!! should go here !  */

  if (!(vcons[vc] = vmalloc(size)))
    return -ENOMEM;			/* seemed to be too much ...	*/

  vcons[vc]->vmalloced = size;		/* we need this later again.	*/
  vc_reset_to_mode(vc, &default_textmode);
  return EOK;
}

static void
vc_free(int vc)
{
  vfree(vcons[vc]);
  vcons[vc] = NULL;
}

/*\
|*|	GGI driver interface
\*/

static int
kgi_driver_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		 unsigned long arg)
{
  struct kgi_state *gr = file->private_data;
  int err;
  int console = (fg_console < MAX_NR_CONSOLES)
                            ? fg_console : fg_console - MAX_NR_CONSOLES;

  switch(cmd)
   {
     case DRIVER_TESTDRIVER:	/* check if driver is present	*/
	return DEVICEFILE_MAGIC;

     case DRIVER_TESTMODE:		/* check if mode is possible	*/
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)	*/
	err = kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE);
	return err;

     case DRIVER_SETGRAPHMODE:
	_out_buf = _io_buf;		/* _io_cnt is sizeof(ggi_mode)  */

	if ((err = kgi_check_mode((struct kgi_mode *) _io_buf, CMD_PROPOSE)))
	  return err;

	gr->mode = (*(struct kgi_mode*) _io_buf);
	if (gr->graph_num == console)
	  kgi_set_mode(&(gr->mode));

	/* !!! buffer reallocation!  */
	return EOK;

     case DRIVER_GETGRAPHMODE:	/* return graphics mode		*/
	_out_buf = &(gr->mode.request);
	return EOK;			/* _io_cnt is sizeof(ggi_mode)	*/

     case DRIVER_GRAPHICS:		/* switch to graphics console	*/
	TRACE(printk("ioctl: switch to graphics: "));
	if (gr->graph_num == console)
	  change_console(console + MAX_NR_CONSOLES);
	
	return EOK;

     case DRIVER_SETTEXTMODE:	/* set (console) text mode	*/
	if (((struct kgi_mode *) _io_buf)->request.graphtype != GT_TEXT)
	  {
	    TRACE(printk("ioctl: set textmode but arg is not a text mode.\n"));
	    return -EINVAL;
          }

	if ((err = kgi_check_mode((struct kgi_mode*) _io_buf, CMD_PROPOSE)))
	  return err;

	vc_reset_to_mode(gr->graph_num, (ggi_mode*) _io_buf);

	/* !!!	buffer reallocation */

	_out_buf = textmode + gr->graph_num;
	return EOK;			/* _io_cnt is sizeof(ggi_mode)	*/
	
     case DRIVER_GETTEXTMODE:	/* return (console) textmode	*/
	_out_buf = textmode + gr->graph_num;
	return EOK;			/* _io_cnt is sizeof(ggi_mode)	*/

     case DRIVER_TEXT:		/* switch to text console	*/
	if (gr->graph_num == console)
	  change_console(console);
	return EOK;			/* sould be always possible */

     default:
	return -EDRIVER;
   }
}

static void
kgi_done(void)
{
  kgi_clock_done();			/* deinitialisation, the calling     */
  kgi_ramdac_done();			/* sequence is reverse to initiali-  */
  kgi_monitor_done();			/* sation.			     */
  kgi_chipset_done();
}

static int
kgi_init(void)
{
  if (kgi_chipset_init())		/* chipset provides IO functions for */
    return -EIO;			/* all other parts, thus init first. */

  if (kgi_monitor_init())		/* chipset ok. Now init the monitor  */
    {					/* driver. If this fails (e.g. not   */
      kgi_chipset_done();		/* detected) deinitialize chipset.   */
      return -EIO;
    }

  if (kgi_ramdac_init())		/* Now go for the ramdac. 	     */
    {					
      kgi_monitor_done();		/* failed, thus deinitialize...	     */
      kgi_chipset_done();
      return -EIO;
    }

  if (kgi_clock_init())			/* last is the clock driver. 	     */
    {
      kgi_ramdac_done();		/* failed :-(			     */
      kgi_monitor_done();
      kgi_chipset_done();
      return -EIO;
    }

  memset(textmode, 0, sizeof(textmode));
  textmode[0].request = default_textmode;	/* check the default textmode */
  if (kgi_check_mode(textmode, CMD_PROPOSE))
    {
      printk("fatal: kgi_check_mode(default_textmode) failed.\n");
      kgi_done();
      return -EIO;
    }

  return 0;
}

/*\
|*|	/dev/graph file operations
\*/


typedef int (*ioctl_fn)(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg); 

static ioctl_fn ioctl[8] =
{
  kgi_driver_ioctl,	/* NOTE: this depends on the IOCTL definitions ! */
  kgi_monitor_ioctl,
  kgi_ramdac_ioctl,
  kgi_clock_ioctl,
  kgi_chipset_ioctl,
  kgi_graphics_ioctl,
  NULL,
  NULL
};

unsigned int	_io_cnt;
unsigned char   _io_buf[_IO_BUF_SIZE];
         void	*_out_buf;

static int
graph_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	  unsigned long arg)
{
  static int lock = 0;
  register int foo;
  int result = 0;

  if (lock) return -EBUSY;	/* ??? could this happen (new ioctl while  */ 
  lock = 1;			/* ??? the old one is not finished ?)	   */

  _out_buf = NULL;
  _io_cnt = _IOC_SIZE(cmd);

  if (cmd & IOC_IN)		/* do we have input ?	*/
    {
      if ((foo = verify_area (VERIFY_READ, (void *) arg, _io_cnt)))
        { result = foo; goto unlock; }

      TRACE(printk("%s: copying %i bytes fromfs\n",__FILE__, _io_cnt));
      memcpy_fromfs (_io_buf, (void *) arg, _io_cnt);
    }

  foo = _IOC_TYPE(cmd);
  if ((foo < 8) && ioctl[foo])
    result = ioctl[foo] (inode, file, cmd, arg);
   else
    {
      TRACE(printk("%s:%i: invalid subsystem (%i).\n",__FILE__, __LINE__, foo));
      result = -EINVAL;  goto unlock;
    }

  if (cmd & IOC_OUT)
    {
      if ((foo = verify_area (VERIFY_WRITE, (void *) arg, _io_cnt)))
        { result = foo; goto unlock; }
      if (_out_buf && (0 < _io_cnt) && (_io_cnt < (IOCSIZE_MASK+1)))
        {
	  TRACE(printk("%s: copying %i bytes tofs\n",__FILE__, _io_cnt));
          memcpy_tofs ((void *) arg, _out_buf, _io_cnt);
        }
       else
        result = -EDRIVER;
    }

  unlock:
    lock = 0;
    TRACE(printk("%s: ioctl() = 0x%.8x\n", __FILE__, result));
    return result;
}

static int 
graph_select(struct inode *inode, struct file *file, int type, select_table *wait)
{
  struct kgi_state *gr = file->private_data;

  switch (type)
   {
     case SEL_IN:	/* read access possible?	*/

	if (can_get_event(&gr->ev_buffer))
          return 1;
        select_wait(&gr->ev_buffer.wait, wait);
	return 0;

     case SEL_OUT:	/* write access possible?	*/
     default:
	return 0;
   }
}

static int 
graph_read(struct inode *inode, struct file *file, char *buf, int count)
{
  struct kgi_state *gr = file->private_data;
  int write = gr->ev_buffer.write;
  int read  = gr->ev_buffer.read;
  int cnt = write - read;
  int err;

  while (! cnt)			/* if there is no input availiable now, */
    {				/* block the process until we have sth. */
      if (current->signal & ~current->blocked)
        return 0;

      current->state = TASK_INTERRUPTIBLE;
      current->timeout = jiffies + 10; /* ms */
      schedule();

      write = gr->ev_buffer.write;
      cnt = write - read;
    }

  count /= sizeof(ggi_event);
  if (count < cnt)		/* more availiable than requested ?	*/
    {
      cnt = count;
      write = read + cnt; 
    }
  if (!cnt)
    return 0;

  if ((err = verify_area (VERIFY_WRITE, buf, cnt*sizeof(ggi_event))))
    return err;    /* ??? is this check neccessary or does VFS aleready ? */

  if (write < gr->ev_buffer.size)	/* can we read all at once ?	*/
    {
      int written = cnt*sizeof(ggi_event);

      memcpy_tofs(buf, &(gr->event[read]), written);	/* read out events */
      gr->ev_buffer.read = write;

      inode->i_atime = CURRENT_TIME;

      return written;
    }
   else
    {
      int first  = sizeof(ggi_event)*(gr->ev_buffer.size - read);
      int second = sizeof(ggi_event)*(write - gr->ev_buffer.size);

      memcpy_tofs(buf, &(gr->event[read]), first);		/* read out events */
      memcpy_tofs(buf+first, &(gr->event[0]), second);
      gr->ev_buffer.write &= gr->ev_buffer.size - 1;
      gr->ev_buffer.read  = write & (gr->ev_buffer.size - 1);

      inode->i_atime = CURRENT_TIME;

      return first + second;
    }
}

static int
graph_open(struct inode *inode, struct file *file)
{
  struct tty_struct *tty = current->tty;
  int minor = MINOR(inode->i_rdev) ? MINOR(inode->i_rdev) : MINOR(tty->device);

  if (   ((tty->driver.type != TTY_DRIVER_TYPE_CONSOLE)
          && (!suser() || MINOR(inode->i_rdev) == 0))
      ||  MINOR(tty->device) != minor) /* !!! change_console depends on this */
    return -EPERM;	/* no way, the controlling tty must be local!	*/

  minor--;

  if (graph[minor])	/* for now we allow only one process per VG	*/
    return -EBUSY;

  if (minor < MAX_NR_CONSOLES)
    {
      struct kgi_state *gr = vmalloc(sizeof(struct kgi_state));

      if (! gr)
        return -ENOMEM;

      file->private_data = graph[minor] = gr;

      memset(gr, 0, sizeof(*gr));

      gr_kbd[minor].ledflagstate = gr_kbd[minor].default_ledflagstate;
      gr_kbd[minor].ledmode      = LED_SHOW_FLAGS;
      gr_kbd[minor].lockstate    = 0;	/* reset keyboard	*/
      gr_kbd[minor].kbdmode      = VC_KGI;
      
      gr->graph_num = minor;
      gr->mode = textmode[minor];	/* default to textmode	*/

      gr->ev_buffer.event = gr->event;	/* setup event buffer	*/
      gr->ev_buffer.size = MAX_EVENTS;
      gr->ev_buffer.selected = emKeyPress | emKeyRepeat;
      ev_buffer[minor] = &gr->ev_buffer;

      return EOK;
    }
  return -ENODEV;
}

static void
graph_close(struct inode *inode, struct file *file)
{
  struct kgi_state *gr = file->private_data;

  graph[gr->graph_num] = NULL;
  kgi_free_state(gr);
}

static struct file_operations 
kgi_fops =
{ NULL,		/* lseek	*/
  graph_read,	/* read		*/
  NULL,         /* write	*/
  NULL,         /* readdir	*/
  graph_select,	/* select	*/
  graph_ioctl,	/* ioctl	*/ 
  NULL,		/* mmap		*/
  graph_open,	/* open		*/
  graph_close,	/* release	*/
  NULL		/* fsync	*/
};

/*\
|*|	console file operation enhancements/replacements
\*/

static void
setup_hardware(int want)
{
  if (fg_console < MAX_NR_CONSOLES)	/* text screen ? */
    {
       struct kgi_vc *cons = vcons[fg_console];
       u_short *d          = (void *)(cons + 1);
       u_short *s          = FB;
       unsigned int count  = FB_END;

       if (CURSOR.hide)		/* undo any modifications (POINTER too !!!) */
         CURSOR.hide(&CURSOR, POS);
       if (POINTER.hide)
         POINTER.hide(&POINTER, PPOS);

       FB = d;			/* This is to be the new framebuffer.	*/
       while (count--)		/* Copy contents from hardware to RAM.	*/
         *(d++) = *(s++);

       SYNC        = NULL;	/* unset all hardware support		*/
       SET_ORIGIN  = NULL;
       CURSOR.hide = POINTER.hide = NULL;
       CURSOR.show = POINTER.show = NULL;
    }
   else
    {
       /* !!!	save graphics screen.	*/
    }

  if (want < MAX_NR_CONSOLES)
    {
       struct kgi_vc *cons = vcons[want];
       u_short *d          = (void *) fb_text;
       u_short *s          = FB;
       unsigned int count  = FB_END;

       kgi_set_mode(textmode + want);

       FB = d;			/* This is to be the new framebuffer.	*/
       while (count--)		/* Copy contents from RAM to hardware.	*/
         *(d++) = *(s++);

       SET_ORIGIN   = textmode[want].set_origin;/* set hardware support	*/
       CURSOR.hide  = textmode[want].cursor_hide;
       CURSOR.show  = textmode[want].cursor_show;
       POINTER.hide = textmode[want].pointer_hide;
       POINTER.show = textmode[want].pointer_show;

       __ORIGIN = ORIGIN;

       SET_ORIGIN(ORIGIN);
       if (CURSOR_MODE)
         if (CURSOR.show)
           CURSOR.show(&CURSOR, POS);

       if (POINTER_MODE)
         if (POINTER.show)
           POINTER.show(&POINTER, PPOS);
           

       /* !!! SYNC too!, also show cursor/pointer! */
    }
   else 
    {
      struct kgi_state *gr = graph[want - MAX_NR_CONSOLES];

      kgi_set_mode(&gr->mode);
      /* !!! mmapping, framebuffer, ... */
    }
}

static struct tty_driver old_driver;

static int
con_open(struct tty_struct *tty, struct file *filp)
{
  unsigned int index  = MINOR(tty->device) - tty->driver.minor_start;
  int err;

  if (!(vcons[index]))			/* not allocated yet ?	*/
    if ((err = vc_allocate(index)))
      return err;

  return old_driver.open(tty, filp);
}

static void
con_close(struct tty_struct *tty, struct file *filp)
{
  struct kgi_vc *cons = tty->driver_data;

  if (tty->count == 1)
      vc_reset_to_mode(VC_NUM, &default_textmode);

  if (old_driver.close)
    old_driver.close(tty, filp);
}

/*\
|*|	module loading
\*/

int
init_module(void)
{
   memset(&old_driver, 0, sizeof(old_driver));
   old_driver.open  = con_open;
   old_driver.close = con_close; 

   if (change_console_tty_ops(&old_driver, setup_hardware))
     {
       printk("%s: can't change console file operations.\n",__FILE__);
       return -EIO;
     }

   if (register_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME, &kgi_fops))
     {
       reset_console_tty_ops();
       printk("%s: can't register device file %s (major = %i).\n",\
	      __FILE__, DEVICEFILE_NAME, DEVICEFILE_MAJOR);
       reset_console_tty_ops();
       return -EIO;
     }

   if(kgi_init())
     {
       unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME);
       reset_console_tty_ops();
       return -EIO;
     }

   if (vc_allocate(0))
     {
        unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME);
        reset_console_tty_ops();
        kgi_done();
        return -EIO;
     }

   old_driver.table[0]->driver_data = vcons[0];	/* we changed vcons[0] ! */

   printk("kgi module loaded.\n");
   return 0;
}

void
cleanup_module (void)
{
  if (MOD_IN_USE)
    printk("kgi: scrdrv module is still in use. (not removed)\n");

  if (unregister_chrdev(DEVICEFILE_MAJOR, DEVICEFILE_NAME))
    printk("kgi: %s:%i: cleanup_module() failed.\n", __FILE__, __LINE__);
   else
    {
      kgi_done();
      printk("kgi module removed.\n");
    }

  /* !!! we have to restore console state... (difficult) ! */
}

