/*\
|*| --------------------------------------------------------------------------
|*|	Generic Monitor driver for fixed frequency monitors
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995, Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#define	MON_VERSION	"0.00 ALPHA"

#ifdef DEBUG
#define NEED_MONITOR_SYMS
#endif

#include "graphdev.h"
#include "kgi.h"

#if !((MONITORVERSION == 01) && (MODEVERSION == 03))
  #warning kgi data structures have changes, check this code !
#endif

#define HPOS            0x1000
#define HNEG            0
#define VPOS            0x2000
#define VNEG            0

#define	MAX_HTIMINGS	16	/* MUST BE A POWER OF 2 !!	*/
#define	HTIMING_MASK	(MAX_HTIMINGS-1)

#include "monitors.inc"

#if HTIMINGS > MAX_HTIMINGS
#error more htimings than allowed by MAX_HTIMINGS
#endif

/*\
|*|	GGI interface
\*/

int
kgi_monitor_init(void)
{
  TRACE(printk("%s %s monitor driver v" MON_VERSION "\n", \
        monitor.manufact, monitor.model));
  TRACE(printk("(%s - %s, %dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)\n",\
	  DISPLAYTYPE(monitor.disptype),\
	  (monitor.disptype & MONO_DISPLAY) ? "monochrome" : "color",\
	  monitor.maxdots.x,\
	  monitor.maxdots.y,\
	  monitor.size.x,\
	  monitor.size.y,\
	  monitor.hfreq.min/1000, monitor.hfreq.max/1000, \
	  monitor.vfreq.min      ,monitor.vfreq.max ));
  return EOK;
}

void
kgi_monitor_done(void)
{
   TRACE(printk("%s %s monitor driver removed.\n", \
         monitor.manufact, monitor.model));
}

struct kgi_monitor
*kgi_monitor_get_info (void)
{
  return &monitor;
}

static int err(int pixels, int ns, struct kgi_mode *mode)
{ int h = (mode->request.graphtype == GT_TEXT)
	   ? mode->request.textgrid.x
	   : 9; 	/* timings may differ by up to 9 dclks */
  int ret = abs(pixels - (int)(ns*(1e-9*mode->dclk))) > h;

  if (ret)
   printk("%s: timing wrong, because %i - %i > %i\n", __FILE__, pixels, 
	  (int)(ns*(1e-9*mode->dclk)), h);

  return ret;
}

int
kgi_monitor_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd *cmd)
{
  int v = 0, h = 0;
  int dclk;

  #define VISX   (mode->x.width)
  #define VISY   (mode->y.width)

  if ((VISX > (int)(monitor.maxdots.x)) || (VISY > (int)(monitor.maxdots.y)))
    {
      TRACE(printk("%s: resolution %ix%i pixels too high.\n",\
                   __FILE__, VISX, VISY));
      return -EMONITOR;
    }

  while ((v < VTIMINGS) && ((int)vtiming[v].width != VISY))
    v++;
  if ((v == VTIMINGS) || (vtiming[v].width != VISY))
    {
      TRACE(printk("%s: no suitable vtiming for %i lines found.\n",\
                   __FILE__, VISY));
      return -EMONITOR;
    }
  h = vtiming[v].polarity & HTIMING_MASK;
  dclk = (int)((mode->x.width*1e9) / htiming[h].width);

  switch (*cmd)
   {
     case CMD_PROPOSE:
	mode->dclk = dclk;
	mode->y    = vtiming[v];
	mode->y.polarity = (vtiming[v].polarity & VPOS) ? 1 : 0;

	#define POS(X)  (((int) VISX * htiming[h].X) / htiming[h].width)
	mode->x.width      = POS(width);
	mode->x.blankstart = POS(blankstart);
	mode->x.syncstart  = POS(syncstart);
	mode->x.syncend    = POS(syncend);
	mode->x.blankend   = POS(blankend);
	mode->x.total      = POS(total);
	mode->x.polarity   = (vtiming[v].polarity & HPOS) ? 1 : 0;
	#undef POS

	if (monitor.flags & FL_MON_PROPSIZE)
	  {
 	    mode->request.size.x = VISX * monitor.size.x / monitor.maxdots.x;
	    mode->request.size.y = VISY * monitor.size.y / monitor.maxdots.y;
          }
	 else
	  mode->request.size = monitor.size;

	TRACE(printk("%s: propose dclk = %i Hz hfreq = %i Hz vfreq = %i Hz\n",\
	             __FILE__, mode->dclk, mode->dclk / mode->x.width, \
	             mode->dclk / (mode->x.total*mode->y.total)));

	*cmd = CMD_LOWER;
	return EOK;

     case CMD_LOWER:
	if (mode->dclk < monitor.dclk.min)
	  return -EMONITOR;

	if (abs(mode->dclk - dclk) > (2*dclk / 100))
	  {
	    TRACE(printk("%s: dclk should be near %i Hz but is %i Hz\n", \
	                 __FILE__, dclk, mode->dclk));
	    return -EMONITOR;
          }

	*cmd = CMD_CHECK;
	return EOK;

     case CMD_RAISE:
	if (mode->dclk > monitor.dclk.max)
	  return -EMONITOR;

	if (abs(mode->dclk - dclk) > (2*dclk / 100))
          {
	    TRACE(printk("%s: dclk should be near %i Hz but is %i Hz\n", \
	                 __FILE__, dclk, mode->dclk));
	    return -EMONITOR;
	  }

	*cmd = CMD_CHECK;
	return EOK;

     case CMD_CHECK:
	if ((mode->dclk < monitor.dclk.min) || (monitor.dclk.max < mode->dclk))
	  {
	     TRACE(printk("%s: dclk of %i Hz is out of bounds.\n",\
	                  __FILE__, mode->dclk));
	     return -EMONITOR;
	  }
	if (   (monitor.maxdots.x < (int) VISX) 
	    || (monitor.maxdots.y < (int) VISY))
	  {
	     TRACE(printk("%s: resolution of %ix%i pixels is too high.\n", \
	                  __FILE__, VISX, VISY));
	     return -EMONITOR;
	  }

	if (   (abs(mode->dclk - dclk) > (2*dclk/100))
	    || ((mode->dclk / mode->x.total) < monitor.hfreq.min)
	    || ((mode->dclk / mode->x.total) > monitor.hfreq.max))
	  {
	    TRACE(printk("%s: hfreq (%i Hz) out of bounds.\n", \
	                 __FILE__, mode->dclk / mode->x.total));
	    return -EMONITOR;
	  }
	
	#define	ERR(X) (err(mode->x.X, htiming[h].X, mode) \
	                || (mode->y.X != vtiming[v].X))
	if (   ERR(width)   || ERR(blankstart) || ERR(syncstart) 
	    || ERR(syncend) || ERR(blankend)   || ERR(total))
	  {
	    TRACE(printk("%s: error in timings.\n", __FILE__));
	    return -EMONITOR;
	  }
	#undef ERR

	*cmd = CMD_READY;
	return EOK;

     default:
	INTERNALERROR;
	return -EMONITOR;
   }

  #undef VISX
  #undef VISY
}

void
kgi_monitor_set_mode(struct kgi_mode *TM)
{
}

int
kgi_monitor_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  switch(cmd)
   {
     case MONITOR_GETINFO:
	_out_buf = &monitor; _io_cnt = sizeof(monitor);
	return EOK;
     case MONITOR_SETINFO:
	if (suser())
	  {
	    memcpy(&monitor, _io_buf, sizeof(monitor));
	    return EOK;
	  }
	 else
	  return -EPERM;
     default:
	return -EMONITOR;
   }
}
